from __future__ import annotations

import abc

import pybamm


class BaseProcessedVariable(abc.ABC):
    """
    Shared API for both 'on-demand' and 'computed' processed variables.
    """

    @abc.abstractmethod
    def __call__(self, **coords): ...

    @abc.abstractmethod
    def as_computed(self) -> pybamm.ProcessedVariableComputed: ...

    def update(
        self,
        other: BaseProcessedVariable,
        new_sol: pybamm.Solution,
    ) -> pybamm.ProcessedVariableComputed:
        """
        Return a *new* CPV that is `self` followed by `other`.
        Works no matter which concrete types we start with.
        """
        return self.as_computed()._concat(other.as_computed(), new_sol)



#
# Class for leading-order electrolyte diffusion employing stefan-maxwell
#
import numpy as np

import pybamm

from .base_electrolyte_diffusion import BaseElectrolyteDiffusion


class LeadingOrder(BaseElectrolyteDiffusion):
    """Class for conservation of mass in the electrolyte employing the
    Stefan-Maxwell constitutive equations. (Leading refers to leading order
    of asymptotic reduction)

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    """

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):
        c_e_av = pybamm.Variable(
            "X-averaged electrolyte concentration [mol.m-3]",
            domain="current collector",
            bounds=(0, np.inf),
        )
        c_e_av.print_name = "c_e_av"

        c_e_dict = {
            domain: pybamm.PrimaryBroadcast(c_e_av, domain)
            for domain in self.options.whole_cell_domains
        }

        variables = self._get_standard_concentration_variables(c_e_dict)

        N_e = pybamm.FullBroadcastToEdges(
            0, self.options.whole_cell_domains, "current collector"
        )
        variables.update(self._get_standard_flux_variables(N_e))

        return variables

    def get_coupled_variables(self, variables):
        c_e = variables["Electrolyte concentration [mol.m-3]"]
        eps = variables["Porosity"]

        variables.update(self._get_total_concentration_electrolyte(eps * c_e))

        return variables

    def set_rhs(self, variables):
        c_e_av = variables["X-averaged electrolyte concentration [mol.m-3]"]

        T_av = variables["X-averaged cell temperature [K]"]

        eps_n_av = variables["X-averaged negative electrode porosity"]
        eps_s_av = variables["X-averaged separator porosity"]
        eps_p_av = variables["X-averaged positive electrode porosity"]

        deps_n_dt_av = variables["X-averaged negative electrode porosity change [s-1]"]
        deps_p_dt_av = variables["X-averaged positive electrode porosity change [s-1]"]

        div_Vbox_s_av = variables[
            "X-averaged separator transverse volume-averaged acceleration [m.s-2]"
        ]

        sum_a_j_n_0 = variables[
            "Sum of x-averaged negative electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]
        sum_a_j_p_0 = variables[
            "Sum of x-averaged positive electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]
        sum_s_j_n_0 = variables[
            "Sum of x-averaged negative electrode electrolyte "
            "reaction source terms [A.m-3]"
        ]
        sum_s_j_p_0 = variables[
            "Sum of x-averaged positive electrode electrolyte "
            "reaction source terms [A.m-3]"
        ]
        source_terms = (
            self.param.n.L
            * (sum_s_j_n_0 - self.param.t_plus(c_e_av, T_av) * sum_a_j_n_0)
            + self.param.p.L
            * (sum_s_j_p_0 - self.param.t_plus(c_e_av, T_av) * sum_a_j_p_0)
        ) / self.param.F

        self.rhs = {
            c_e_av: 1
            / (
                self.param.n.L * eps_n_av
                + self.param.s.L * eps_s_av
                + self.param.p.L * eps_p_av
            )
            * (
                source_terms
                - c_e_av
                * (self.param.n.L * deps_n_dt_av + self.param.p.L * deps_p_dt_av)
                - c_e_av * self.param.s.L * div_Vbox_s_av
            )
        }

    def set_initial_conditions(self, variables):
        c_e = variables["X-averaged electrolyte concentration [mol.m-3]"]
        self.initial_conditions = {c_e: self.param.c_e_init}



#
# Dummy solver class, for empty models
#
import numpy as np

import pybamm


class DummySolver(pybamm.BaseSolver):
    """Dummy solver class for empty models."""

    def __init__(self):
        super().__init__()
        self.name = "Dummy solver"

    def _integrate(self, model, t_eval, inputs_dict=None, t_interp=None):
        """
        Solve an empty model.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : :class:`numpy.array`, size (k,)
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving

        Returns
        -------
        :class:`pybamm.Solution`
            A Solution object containing the times and values of the solution,
            as well as various diagnostic messages.

        """
        y_sol = np.zeros((1, t_eval.size))
        sol = pybamm.Solution(
            t_eval, y_sol, model, inputs_dict, termination="final time"
        )
        sol.integration_time = 0
        return sol



#
# Base parameters class
#
import warnings

import pybamm


class BaseParameters:
    """
    Overload the `__setattr__` method to record what the variable was called.
    """

    def __getattribute__(self, name):
        """
        Raise more informative error to users when they try to access a
        non-existent attribute, which may have recently changed name
        """
        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            if name == "cap_init":
                warnings.warn(
                    "Parameter 'cap_init' has been renamed to 'Q_init'", stacklevel=2
                )
                return self.Q_init
            for domain in ["n", "s", "p"]:
                if f"_{domain}_" in name or name.endswith(f"_{domain}"):
                    name_without_domain = name.replace(f"_{domain}_", "_").replace(
                        f"_{domain}", ""
                    )
                    if hasattr(self, domain):
                        self_domain = getattr(self, domain)
                        if hasattr(self_domain, name_without_domain):
                            raise AttributeError(
                                f"param.{name} does not exist. It has been renamed to "
                                f"param.{domain}.{name_without_domain}"
                            ) from e
                        elif hasattr(self_domain, "prim") and hasattr(
                            self_domain.prim, name_without_domain
                        ):
                            raise AttributeError(
                                f"param.{name} does not exist. It has been renamed to "
                                f"param.{domain}.prim.{name_without_domain}"
                            ) from e
                        else:
                            raise e
                    else:
                        raise e
            raise e

    def __setattr__(self, name, value):
        if hasattr(self, "domain"):
            d = self.domain[0]
            print_name = f"{name}_{d}"
        else:
            print_name = name
        if isinstance(value, pybamm.Symbol):
            value.print_name = print_name
        super().__setattr__(name, value)

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, extra_options):
        if extra_options is None or type(extra_options) == dict:  # noqa: E721
            self._options = pybamm.BatteryModelOptions(extra_options)
        else:
            self._options = extra_options

    @property
    def domain(self):
        return self._domain

    @domain.setter
    def domain(self, domain):
        self._domain = domain
        if domain is not None:
            self._Domain = domain.capitalize()

    @property
    def domain_Domain(self):
        return self.domain, self._Domain

    def set_phase_name(self):
        if (
            self.phase == "primary"
            and getattr(self.main_param.options, self.domain)["particle phases"] == "1"
        ):
            # Only one phase, no need to distinguish between
            # "primary" and "secondary"
            self.phase_name = ""
            self.phase_prefactor = ""
        else:
            # add a space so that we can use "" or (e.g.) "primary " interchangeably
            # when naming variables
            self.phase_name = self.phase + " "
            self.phase_prefactor = self.phase.capitalize() + ": "


class NullParameters:
    def __getattribute__(self, name):
        "Returns 0 for some parameters that aren't found by __getattribute__"
        if name in ["epsilon_s", "Q_init", "n_Li_init", "Q_Li_init", "R_typ"]:
            return pybamm.Scalar(0)
        else:
            return super().__getattribute__(name)

    def _set_parameters(self):
        pass



from __future__ import annotations

import importlib
import inspect
import json
import numbers
import re
from datetime import datetime
from enum import Enum
from pathlib import Path

import black
import numpy as np

import pybamm

SUPPORTED_SCHEMA_VERSION = "1.0"


class ExpressionFunctionParameter(pybamm.UnaryOperator):
    def __init__(self, name, child, func_name, func_args):
        super().__init__(name, child)
        self.func_name = func_name
        self.func_args = func_args

    def _unary_evaluate(self, child):
        """Evaluate the symbolic expression (the child)"""
        return child

    def to_source(self):
        """
        Creates python source code for the function.
        """
        src = f"def {self.func_name}({', '.join(self.func_args)}):\n"

        # Fix printing of parameters so they print as Parameter('name'). Do this on a
        # copy to avoid modifying the original expression.
        expression = self.child.create_copy()
        for child in expression.pre_order():
            if isinstance(child, pybamm.FunctionParameter):
                # Replace FunctionParameter with a constructor call
                # Build the inputs dict string mapping input names to actual parameter
                # names
                inputs_str = ", ".join(
                    f'"{input_name}": {child.children[i].name}'
                    for i, input_name in enumerate(child.input_names)
                )
                child.print_name = (
                    f'FunctionParameter("{child.name}", {{{inputs_str}}})'
                )
            elif (
                isinstance(child, pybamm.Parameter) and child.name not in self.func_args
            ):
                child.name = f'Parameter("{child.name}")'

        src += f"    return {expression.to_equation()}"

        formatted_src = black.format_str(src, mode=black.FileMode())
        return formatted_src


class Serialise:
    """
    Converts a discretised model to and from a JSON file.

    """

    def __init__(self):
        pass

    class _SymbolEncoder(json.JSONEncoder):
        """Converts PyBaMM symbols into a JSON-serialisable format"""

        def default(self, node: dict):
            node_dict = {"py/object": str(type(node))[8:-2], "py/id": id(node)}
            if isinstance(node, pybamm.Symbol):
                node_dict.update(node.to_json())  # this doesn't include children
                node_dict["children"] = []
                for c in node.children:
                    node_dict["children"].append(self.default(c))

                if hasattr(node, "initial_condition"):  # for ExplicitTimeIntegral
                    node_dict["initial_condition"] = self.default(
                        node.initial_condition
                    )

                return node_dict

            if isinstance(node, pybamm.Event):
                node_dict.update(node.to_json())
                node_dict["expression"] = self.default(node._expression)
                return node_dict

            node_dict["json"] = json.JSONEncoder.default(self, node)  # pragma: no cover
            return node_dict  # pragma: no cover

    class _MeshEncoder(json.JSONEncoder):
        """Converts PyBaMM meshes into a JSON-serialisable format"""

        def default(self, node: pybamm.Mesh):
            node_dict = {"py/object": str(type(node))[8:-2], "py/id": id(node)}
            if isinstance(node, pybamm.Mesh):
                node_dict.update(node.to_json())

                submeshes = {}
                for k, v in node.items():
                    if len(k) == 1 and "ghost cell" not in k[0]:
                        submeshes[k[0]] = self.default(v)

                node_dict["sub_meshes"] = submeshes

                return node_dict

            if isinstance(node, pybamm.SubMesh):
                node_dict.update(node.to_json())
                return node_dict

            node_dict["json"] = json.JSONEncoder.default(self, node)  # pragma: no cover
            return node_dict  # pragma: no cover

    class _Empty:
        """A dummy class to aid deserialisation"""

        pass

    class _EmptyDict(dict):
        """A dummy dictionary class to aid deserialisation"""

        pass

    def serialise_model(
        self,
        model: pybamm.BaseModel,
        mesh: pybamm.Mesh | None = None,
        variables: pybamm.FuzzyDict | None = None,
    ) -> dict:
        """Converts a discretised model to a JSON-serialisable dictionary.

        As the model is discretised and ready to solve, only the right hand side,
        algebraic and initial condition variables are serialised.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The discretised model to be serialised
        mesh : :class:`pybamm.Mesh` (optional)
            The mesh the model has been discretised over. Not necessary to solve
            the model when read in, but required to use pybamm's plotting tools.
        variables: :class:`pybamm.FuzzyDict` (optional)
            The discretised model variables. Not necessary to solve a model, but
            required to use pybamm's plotting tools.

        Returns
        -------
        dict
            A JSON-serialisable dictionary representation of the model
        """
        if model.is_discretised is False:
            raise NotImplementedError(
                "PyBaMM can only serialise a discretised, ready-to-solve model."
            )

        model_json = {
            "py/object": str(type(model))[8:-2],
            "py/id": id(model),
            "pybamm_version": pybamm.__version__,
            "name": model.name,
            "options": model.options,
            "bounds": [bound.tolist() for bound in model.bounds],  # type: ignore[attr-defined]
            "concatenated_rhs": self._SymbolEncoder().default(model._concatenated_rhs),
            "concatenated_algebraic": self._SymbolEncoder().default(
                model._concatenated_algebraic
            ),
            "concatenated_initial_conditions": self._SymbolEncoder().default(
                model._concatenated_initial_conditions
            ),
            "events": [self._SymbolEncoder().default(event) for event in model.events],
            "mass_matrix": self._SymbolEncoder().default(model.mass_matrix),
            "mass_matrix_inv": self._SymbolEncoder().default(model.mass_matrix_inv),
        }

        if mesh:
            model_json["mesh"] = self._MeshEncoder().default(mesh)

        if variables:
            if model._geometry:
                model_json["geometry"] = self._deconstruct_pybamm_dicts(model._geometry)
            model_json["variables"] = {
                k: self._SymbolEncoder().default(v) for k, v in dict(variables).items()
            }

        return model_json

    def save_model(
        self,
        model: pybamm.BaseModel,
        mesh: pybamm.Mesh | None = None,
        variables: pybamm.FuzzyDict | None = None,
        filename: str | None = None,
    ):
        """Saves a discretised model to a JSON file.

        As the model is discretised and ready to solve, only the right hand side,
        algebraic and initial condition variables are saved.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The discretised model to be saved
        mesh : :class:`pybamm.Mesh` (optional)
            The mesh the model has been discretised over. Not neccesary to solve
            the model when read in, but required to use pybamm's plotting tools.
        variables: :class:`pybamm.FuzzyDict` (optional)
            The discretised model varaibles. Not necessary to solve a model, but
            required to use pybamm's plotting tools.
        filename: str (optional)
            The desired name of the JSON file. If no name is provided, one will be
            created based on the model name, and the current datetime.
        """
        model_json = self.serialise_model(model, mesh, variables)

        if filename is None:
            filename = model.name + "_" + datetime.now().strftime("%Y_%m_%d-%p%I_%M")

        with open(filename + ".json", "w") as f:
            json.dump(model_json, f)

    def load_model(
        self, filename: str | dict, battery_model: pybamm.BaseModel | None = None
    ) -> pybamm.BaseModel:
        """
        Loads a discretised, ready to solve model into PyBaMM.

        A new pybamm battery model instance will be created, which can be solved
        and the results plotted as usual.

        Currently only available for pybamm models which have previously been written
        out using the `save_model()` option.

        Warning: This only loads in discretised models. If you wish to make edits to the
        model or initial conditions, a new model will need to be constructed seperately.

        Parameters
        ----------

        filename: str or dict
            Path to the JSON file containing the serialised model file, or a dictionary
            containing the serialised model data
        battery_model:  :class:`pybamm.BaseModel` (optional)
            PyBaMM model to be created (e.g. pybamm.lithium_ion.SPM), which will
            override any model names within the file. If None, the function will look
            for the saved object path, present if the original model came from PyBaMM.

        Returns
        -------
        :class:`pybamm.BaseModel`
            A PyBaMM model object, of type specified either in the JSON or in
            `battery_model`.
        """

        if isinstance(filename, dict):
            model_data = filename
        else:
            with open(filename) as f:
                model_data = json.load(f)

        recon_model_dict = {
            "name": model_data["name"],
            "options": self._convert_options(model_data["options"]),
            "bounds": tuple(np.array(bound) for bound in model_data["bounds"]),
            "concatenated_rhs": self._reconstruct_expression_tree(
                model_data["concatenated_rhs"]
            ),
            "concatenated_algebraic": self._reconstruct_expression_tree(
                model_data["concatenated_algebraic"]
            ),
            "concatenated_initial_conditions": self._reconstruct_expression_tree(
                model_data["concatenated_initial_conditions"]
            ),
            "events": [
                self._reconstruct_expression_tree(event)
                for event in model_data["events"]
            ],
            "mass_matrix": self._reconstruct_expression_tree(model_data["mass_matrix"]),
            "mass_matrix_inv": self._reconstruct_expression_tree(
                model_data["mass_matrix_inv"]
            ),
        }

        recon_model_dict["geometry"] = (
            self._reconstruct_pybamm_dict(model_data["geometry"])
            if "geometry" in model_data.keys()
            else None
        )

        recon_model_dict["mesh"] = (
            self._reconstruct_mesh(model_data["mesh"])
            if "mesh" in model_data.keys()
            else None
        )

        recon_model_dict["variables"] = (
            {
                k: self._reconstruct_expression_tree(v)
                for k, v in model_data["variables"].items()
            }
            if "variables" in model_data.keys()
            else None
        )

        if battery_model:
            return battery_model.deserialise(recon_model_dict)

        if "py/object" in model_data.keys():
            model_framework = self._get_pybamm_class(model_data)
            return model_framework.deserialise(recon_model_dict)

        raise TypeError(
            """
            The PyBaMM battery model to use has not been provided.
            """
        )

    @staticmethod
    def _json_encoder(obj):
        if isinstance(obj, Enum):
            return obj.name
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.integer):
            return int(obj)
        else:
            raise TypeError(f"Object of type {type(obj)} is not JSON serializable.")

    @staticmethod
    def serialise_custom_model(model: pybamm.BaseModel) -> dict:
        """
        Converts a custom (non-discretised) PyBaMM model to a JSON-serialisable dictionary.

        This includes symbolic expressions for rhs, algebraic, initial and boundary
        conditions, events, and variables. Works for user defined models that are
        subclasses of BaseModel.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The custom symbolic model to be serialised.

        Returns
        -------
        dict
            A JSON-serialisable dictionary representation of the model

        Raises
        ------
        AttributeError
            If the model is missing required sections
        """
        required_attrs = [
            "rhs",
            "algebraic",
            "initial_conditions",
            "boundary_conditions",
            "events",
            "variables",
        ]
        missing = [attr for attr in required_attrs if not hasattr(model, attr)]
        if missing:
            raise AttributeError(f"Model is missing required sections: {missing}")

        base_cls = model.__class__.__bases__[0] if model.__class__.__bases__ else object
        # If the base class is object or builtins.object, use pybamm.BaseModel instead
        if base_cls is object or (
            base_cls.__module__ == "builtins" and base_cls.__name__ == "object"
        ):
            base_cls_str = "pybamm.BaseModel"
        else:
            base_cls_str = f"{base_cls.__module__}.{base_cls.__name__}"

        model_content = {
            "name": getattr(model, "name", "unnamed_model"),
            "base_class": base_cls_str,
            "options": getattr(model, "options", {}),
            "rhs": [
                (
                    convert_symbol_to_json(variable),
                    convert_symbol_to_json(rhs_expression),
                )
                for variable, rhs_expression in getattr(model, "rhs", {}).items()
            ],
            "algebraic": [
                (
                    convert_symbol_to_json(variable),
                    convert_symbol_to_json(algebraic_expression),
                )
                for variable, algebraic_expression in getattr(
                    model, "algebraic", {}
                ).items()
            ],
            "initial_conditions": [
                (
                    convert_symbol_to_json(variable),
                    convert_symbol_to_json(initial_value),
                )
                for variable, initial_value in getattr(
                    model, "initial_conditions", {}
                ).items()
            ],
            "boundary_conditions": [
                (
                    convert_symbol_to_json(variable),
                    {
                        side: [
                            convert_symbol_to_json(expression),
                            boundary_type,
                        ]
                        for side, (expression, boundary_type) in conditions.items()
                    },
                )
                for variable, conditions in getattr(
                    model, "boundary_conditions", {}
                ).items()
            ],
            "events": [
                {
                    "name": event.name,
                    "expression": convert_symbol_to_json(event.expression),
                    "event_type": event.event_type,
                }
                for event in getattr(model, "events", [])
            ],
            "variables": {
                str(variable_name): convert_symbol_to_json(expression)
                for variable_name, expression in getattr(model, "variables", {}).items()
            },
        }

        SCHEMA_VERSION = "1.0"
        model_json = {
            "schema_version": SCHEMA_VERSION,
            "pybamm_version": pybamm.__version__,
            "model": model_content,
        }

        return model_json

    @staticmethod
    def save_custom_model(
        model: pybamm.BaseModel, filename: str | Path | None = None
    ) -> None:
        """
        Saves a custom (non-discretised) PyBaMM model to a JSON file. Works for user defined models that are subclasses of BaseModel.

        This includes symbolic expressions for rhs, algebraic, initial and boundary
        conditions, events, and variables. Useful for storing or sharing models
        before discretisation.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The custom symbolic model to be saved.
        filename : str, optional
            The desired name of the JSON file. If not provided, a name will be
            generated from the model name and current datetime.

        Example
        -------
        >>> import pybamm
        >>> model = pybamm.lithium_ion.BasicDFN()
        >>> from pybamm.expression_tree.operations.serialise import Serialise
        >>> Serialise.save_custom_model(model, "basicdfn_model.json")

        """
        try:
            model_json = Serialise.serialise_custom_model(model)

            # Extract model name for filename generation
            model_name = model_json["model"]["name"]

            if filename is None:
                safe_name = re.sub(r"[^\w\-_.]", "_", model_name or "unnamed_model")
                timestamp = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
                filename = f"{safe_name}_{timestamp}.json"
                filename = Path(filename)
            else:
                filename = Path(filename)

                if not filename.name.endswith(".json"):
                    raise ValueError(
                        f"Filename '{filename}' must end with '.json' extension."
                    )

                # Sanitize only the filename, not the directory path
                safe_stem = re.sub(r"[^\w\-_.]", "_", filename.stem)
                filename = filename.with_name(f"{safe_stem}.json")

            try:
                with open(filename, "w") as f:
                    json.dump(model_json, f, indent=2, default=Serialise._json_encoder)
            except OSError as file_err:
                raise OSError(
                    f"Failed to write model JSON to file '{filename}': {file_err}"
                ) from file_err

        except AttributeError:
            # Let AttributeError propagate directly
            raise
        except Exception as e:
            raise ValueError(f"Failed to save custom model: {e}") from e

    @staticmethod
    def serialise_custom_geometry(geometry: pybamm.Geometry) -> dict:
        """
        Converts a custom PyBaMM geometry to a JSON-serialisable dictionary.

        Parameters
        ----------
        geometry : :class:`pybamm.Geometry`
            The geometry object to be serialised.

        Returns
        -------
        dict
            A JSON-serialisable dictionary representation of the geometry
        """
        # Serialize the geometry dict using convert_symbol_to_json for nested symbols
        geometry_dict_serialized: dict = {}
        for domain, domain_geom in geometry.items():
            geometry_dict_serialized[domain] = {}
            for key, value in domain_geom.items():
                # Convert SpatialVariable keys to strings and serialize the key itself
                if isinstance(key, pybamm.Symbol):
                    key_str = key.name if hasattr(key, "name") else str(key)
                    geometry_dict_serialized[domain]["symbol_" + key_str] = (
                        convert_symbol_to_json(key)
                    )
                    # Serialize the value dict
                    serialized_value = {}
                    for k, v in value.items():
                        if isinstance(v, pybamm.Symbol):
                            serialized_value[k] = convert_symbol_to_json(v)
                        else:
                            serialized_value[k] = v
                    geometry_dict_serialized[domain][key_str] = serialized_value
                elif isinstance(key, str):
                    # String keys (like 'tabs') - keep as is
                    if isinstance(value, dict):
                        serialized_value = {}
                        for k, v in value.items():
                            if isinstance(v, pybamm.Symbol):
                                serialized_value[k] = convert_symbol_to_json(v)
                            else:
                                serialized_value[k] = v
                        geometry_dict_serialized[domain][key] = serialized_value
                    else:
                        geometry_dict_serialized[domain][key] = value

        SCHEMA_VERSION = "1.0"
        geometry_json = {
            "schema_version": SCHEMA_VERSION,
            "pybamm_version": pybamm.__version__,
            "geometry": geometry_dict_serialized,
        }

        return geometry_json

    @staticmethod
    def save_custom_geometry(
        geometry: pybamm.Geometry, filename: str | Path | None = None
    ) -> None:
        """
        Saves a custom PyBaMM geometry to a JSON file.

        Parameters
        ----------
        geometry : :class:`pybamm.Geometry`
            The geometry object to be saved.
        filename : str or Path, optional
            The desired name of the JSON file. If not provided, a name will be
            generated using current datetime.
        """
        try:
            geometry_json = Serialise.serialise_custom_geometry(geometry)

            if filename is None:
                timestamp = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
                filename = f"geometry_{timestamp}.json"
                filename = Path(filename)
            else:
                filename = Path(filename)

                if not filename.name.endswith(".json"):
                    raise ValueError(
                        f"Filename '{filename}' must end with '.json' extension."
                    )

                # Sanitize only the filename, not the directory path
                safe_stem = re.sub(r"[^\w\-_.]", "_", filename.stem)
                filename = filename.with_name(f"{safe_stem}.json")

            try:
                with open(filename, "w") as f:
                    json.dump(
                        geometry_json, f, indent=2, default=Serialise._json_encoder
                    )
            except OSError as file_err:
                raise OSError(
                    f"Failed to write geometry JSON to file '{filename}': {file_err}"
                ) from file_err

        except Exception as e:
            raise ValueError(f"Failed to save custom geometry: {e}") from e

    @staticmethod
    def load_custom_geometry(filename: str | dict) -> pybamm.Geometry:
        """
        Loads a custom PyBaMM geometry from a JSON file or dictionary.

        Parameters
        ----------
        filename : str or dict
            Path to the JSON file containing the saved geometry, or a dictionary
            containing the serialised geometry data.

        Returns
        -------
        :class:`pybamm.Geometry`
            The reconstructed geometry object.
        """
        if isinstance(filename, dict):
            data = filename
        else:
            try:
                with open(filename) as file:
                    data = json.load(file)
            except FileNotFoundError as err:
                raise FileNotFoundError(f"Could not find file: {filename}") from err
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"The file '{filename}' contains invalid JSON: {e!s}"
                ) from e

        # Validate schema version
        schema_version = data.get("schema_version", SUPPORTED_SCHEMA_VERSION)
        if schema_version != SUPPORTED_SCHEMA_VERSION:
            raise ValueError(
                f"Unsupported schema version: {schema_version}. "
                f"Expected: {SUPPORTED_SCHEMA_VERSION}"
            )

        # Extract geometry data
        geometry_data = data.get("geometry")
        if geometry_data is None:
            raise KeyError("Missing 'geometry' section in JSON data.")

        # Reconstruct geometry
        reconstructed_geometry: dict = {}

        for domain, domain_geom in geometry_data.items():
            reconstructed_geometry[domain] = {}

            # Find symbol keys and reconstruct SpatialVariables
            symbol_keys = {}
            for key in domain_geom.keys():
                if key.startswith("symbol_"):
                    var_name = key[7:]  # Remove "symbol_" prefix
                    symbol_keys[var_name] = convert_symbol_from_json(domain_geom[key])

            # Now reconstruct the domain geometry with proper keys
            for key, value in domain_geom.items():
                if key.startswith("symbol_"):
                    continue  # Skip symbol definitions

                if key in symbol_keys:
                    # Use the reconstructed SpatialVariable as key
                    spatial_var = symbol_keys[key]
                    reconstructed_value = {}
                    for k, v in value.items():
                        if isinstance(v, dict) and "type" in v:
                            # Reconstruct PyBaMM Symbol using convert_symbol_from_json
                            reconstructed_value[k] = convert_symbol_from_json(v)
                        else:
                            reconstructed_value[k] = v
                    reconstructed_geometry[domain][spatial_var] = reconstructed_value
                else:
                    # String key (like 'tabs')
                    if isinstance(value, dict):
                        reconstructed_value = {}
                        for k, v in value.items():
                            if isinstance(v, dict) and "type" in v:
                                reconstructed_value[k] = convert_symbol_from_json(v)
                            else:
                                reconstructed_value[k] = v
                        reconstructed_geometry[domain][key] = reconstructed_value
                    else:
                        reconstructed_geometry[domain][key] = value

        return pybamm.Geometry(reconstructed_geometry)

    @staticmethod
    def serialise_spatial_methods(spatial_methods: dict) -> dict:
        """
        Converts a dictionary of spatial methods to a JSON-serialisable dictionary.

        Parameters
        ----------
        spatial_methods : dict
            Dictionary mapping domain names to spatial method instances.

        Returns
        -------
        dict
            A JSON-serialisable dictionary representation of the spatial methods
        """
        spatial_methods_dict = {}
        for domain, method in spatial_methods.items():
            spatial_methods_dict[domain] = {
                "class": type(method).__name__,
                "module": type(method).__module__,
                "options": method.options if hasattr(method, "options") else {},
            }

        SCHEMA_VERSION = "1.0"
        spatial_methods_json = {
            "schema_version": SCHEMA_VERSION,
            "pybamm_version": pybamm.__version__,
            "spatial_methods": spatial_methods_dict,
        }

        return spatial_methods_json

    @staticmethod
    def save_spatial_methods(
        spatial_methods: dict, filename: str | Path | None = None
    ) -> None:
        """
        Saves spatial methods to a JSON file.

        Parameters
        ----------
        spatial_methods : dict
            Dictionary mapping domain names to spatial method instances.
        filename : str or Path, optional
            The desired name of the JSON file. If not provided, a name will be
            generated using current datetime.
        """
        try:
            spatial_methods_json = Serialise.serialise_spatial_methods(spatial_methods)

            if filename is None:
                timestamp = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
                filename = f"spatial_methods_{timestamp}.json"
                filename = Path(filename)
            else:
                filename = Path(filename)

                if not filename.name.endswith(".json"):
                    raise ValueError(
                        f"Filename '{filename}' must end with '.json' extension."
                    )

                # Sanitize only the filename, not the directory path
                safe_stem = re.sub(r"[^\w\-_.]", "_", filename.stem)
                filename = filename.with_name(f"{safe_stem}.json")

            try:
                with open(filename, "w") as f:
                    json.dump(
                        spatial_methods_json,
                        f,
                        indent=2,
                        default=Serialise._json_encoder,
                    )
            except OSError as file_err:
                raise OSError(
                    f"Failed to write spatial methods JSON to file '{filename}': {file_err}"
                ) from file_err

        except Exception as e:
            raise ValueError(f"Failed to save spatial methods: {e}") from e

    @staticmethod
    def load_spatial_methods(filename: str | dict) -> dict:
        """
        Loads spatial methods from a JSON file or dictionary.

        Parameters
        ----------
        filename : str or dict
            Path to the JSON file containing the saved spatial methods, or a dictionary
            containing the serialised spatial methods data.

        Returns
        -------
        dict
            Dictionary mapping domain names to spatial method instances.
        """
        if isinstance(filename, dict):
            data = filename
        else:
            try:
                with open(filename) as file:
                    data = json.load(file)
            except FileNotFoundError as err:
                raise FileNotFoundError(f"Could not find file: {filename}") from err
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"The file '{filename}' contains invalid JSON: {e!s}"
                ) from e

        # Validate schema version
        schema_version = data.get("schema_version", SUPPORTED_SCHEMA_VERSION)
        if schema_version != SUPPORTED_SCHEMA_VERSION:
            raise ValueError(
                f"Unsupported schema version: {schema_version}. "
                f"Expected: {SUPPORTED_SCHEMA_VERSION}"
            )

        # Extract spatial methods data
        spatial_methods_data = data.get("spatial_methods")
        if spatial_methods_data is None:
            raise KeyError("Missing 'spatial_methods' section in JSON data.")

        # Reconstruct spatial methods
        reconstructed_methods = {}
        for domain, method_info in spatial_methods_data.items():
            try:
                module_name = method_info["module"]
                class_name = method_info["class"]
                options = method_info.get("options", {})

                # Import module and get class
                module = importlib.import_module(module_name)
                method_class = getattr(module, class_name)

                # Instantiate with options
                reconstructed_methods[domain] = method_class(options=options)

            except (ModuleNotFoundError, AttributeError) as e:
                raise ImportError(
                    f"Could not import spatial method '{class_name}' from '{module_name}': {e}"
                ) from e
            except Exception as e:
                raise ValueError(
                    f"Failed to reconstruct spatial method for domain '{domain}': {e}"
                ) from e

        return reconstructed_methods

    @staticmethod
    def serialise_var_pts(var_pts: dict) -> dict:
        """
        Converts a var_pts dictionary to a JSON-serialisable dictionary.

        Parameters
        ----------
        var_pts : dict
            Dictionary mapping spatial variable names (str or SpatialVariable) to
            number of points (int).

        Returns
        -------
        dict
            A JSON-serialisable dictionary representation of var_pts
        """
        # Convert all keys to strings
        var_pts_dict = {}
        for key, value in var_pts.items():
            if isinstance(key, str):
                var_pts_dict[key] = value
            elif hasattr(key, "name"):
                # SpatialVariable or similar object with name attribute
                var_pts_dict[key.name] = value
            else:
                raise ValueError(f"Unexpected key type in var_pts: {type(key)}")

        SCHEMA_VERSION = "1.0"
        var_pts_json = {
            "schema_version": SCHEMA_VERSION,
            "pybamm_version": pybamm.__version__,
            "var_pts": var_pts_dict,
        }

        return var_pts_json

    @staticmethod
    def save_var_pts(var_pts: dict, filename: str | Path | None = None) -> None:
        """
        Saves var_pts to a JSON file.

        Parameters
        ----------
        var_pts : dict
            Dictionary mapping spatial variable names to number of points.
        filename : str or Path, optional
            The desired name of the JSON file. If not provided, a name will be
            generated using current datetime.
        """
        try:
            var_pts_json = Serialise.serialise_var_pts(var_pts)

            if filename is None:
                timestamp = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
                filename = f"var_pts_{timestamp}.json"
                filename = Path(filename)
            else:
                filename = Path(filename)

                if not filename.name.endswith(".json"):
                    raise ValueError(
                        f"Filename '{filename}' must end with '.json' extension."
                    )

                # Sanitize only the filename, not the directory path
                safe_stem = re.sub(r"[^\w\-_.]", "_", filename.stem)
                filename = filename.with_name(f"{safe_stem}.json")

            try:
                with open(filename, "w") as f:
                    json.dump(
                        var_pts_json, f, indent=2, default=Serialise._json_encoder
                    )
            except OSError as file_err:
                raise OSError(
                    f"Failed to write var_pts JSON to file '{filename}': {file_err}"
                ) from file_err

        except Exception as e:
            raise ValueError(f"Failed to save var_pts: {e}") from e

    @staticmethod
    def load_var_pts(filename: str | dict) -> dict:
        """
        Loads var_pts from a JSON file or dictionary.

        Parameters
        ----------
        filename : str or dict
            Path to the JSON file containing the saved var_pts, or a dictionary
            containing the serialised var_pts data.

        Returns
        -------
        dict
            Dictionary mapping spatial variable names (strings) to number of points.
        """
        if isinstance(filename, dict):
            data = filename
        else:
            try:
                with open(filename) as file:
                    data = json.load(file)
            except FileNotFoundError as err:
                raise FileNotFoundError(f"Could not find file: {filename}") from err
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"The file '{filename}' contains invalid JSON: {e!s}"
                ) from e

        # Validate schema version
        schema_version = data.get("schema_version", SUPPORTED_SCHEMA_VERSION)
        if schema_version != SUPPORTED_SCHEMA_VERSION:
            raise ValueError(
                f"Unsupported schema version: {schema_version}. "
                f"Expected: {SUPPORTED_SCHEMA_VERSION}"
            )

        # Extract var_pts data
        var_pts_data = data.get("var_pts")
        if var_pts_data is None:
            raise KeyError("Missing 'var_pts' section in JSON data.")

        return var_pts_data

    @staticmethod
    def serialise_submesh_types(submesh_types: dict) -> dict:
        """
        Converts a dictionary of submesh types to a JSON-serialisable dictionary.

        Parameters
        ----------
        submesh_types : dict
            Dictionary mapping domain names to submesh classes or MeshGenerator objects.

        Returns
        -------
        dict
            A JSON-serialisable dictionary representation of the submesh types
        """
        submesh_types_dict = {}
        for domain, submesh_item in submesh_types.items():
            # Handle MeshGenerator wrapper objects
            if hasattr(submesh_item, "submesh_type"):
                submesh_class = submesh_item.submesh_type
            else:
                submesh_class = submesh_item

            submesh_types_dict[domain] = {
                "class": submesh_class.__name__,
                "module": submesh_class.__module__,
            }

        SCHEMA_VERSION = "1.0"
        submesh_types_json = {
            "schema_version": SCHEMA_VERSION,
            "pybamm_version": pybamm.__version__,
            "submesh_types": submesh_types_dict,
        }

        return submesh_types_json

    @staticmethod
    def save_submesh_types(
        submesh_types: dict, filename: str | Path | None = None
    ) -> None:
        """
        Saves submesh types to a JSON file.

        Parameters
        ----------
        submesh_types : dict
            Dictionary mapping domain names to submesh classes.
        filename : str or Path, optional
            The desired name of the JSON file. If not provided, a name will be
            generated using current datetime.
        """
        try:
            submesh_types_json = Serialise.serialise_submesh_types(submesh_types)

            if filename is None:
                timestamp = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
                filename = f"submesh_types_{timestamp}.json"
                filename = Path(filename)
            else:
                filename = Path(filename)

                if not filename.name.endswith(".json"):
                    raise ValueError(
                        f"Filename '{filename}' must end with '.json' extension."
                    )

                # Sanitize only the filename, not the directory path
                safe_stem = re.sub(r"[^\w\-_.]", "_", filename.stem)
                filename = filename.with_name(f"{safe_stem}.json")

            try:
                with open(filename, "w") as f:
                    json.dump(
                        submesh_types_json, f, indent=2, default=Serialise._json_encoder
                    )
            except OSError as file_err:
                raise OSError(
                    f"Failed to write submesh types JSON to file '{filename}': {file_err}"
                ) from file_err

        except Exception as e:
            raise ValueError(f"Failed to save submesh types: {e}") from e

    @staticmethod
    def load_submesh_types(filename: str | dict) -> dict:
        """
        Loads submesh types from a JSON file or dictionary.

        Parameters
        ----------
        filename : str or dict
            Path to the JSON file containing the saved submesh types, or a dictionary
            containing the serialised submesh types data.

        Returns
        -------
        dict
            Dictionary mapping domain names to MeshGenerator objects.
        """
        if isinstance(filename, dict):
            data = filename
        else:
            try:
                with open(filename) as file:
                    data = json.load(file)
            except FileNotFoundError as err:
                raise FileNotFoundError(f"Could not find file: {filename}") from err
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"The file '{filename}' contains invalid JSON: {e!s}"
                ) from e

        # Validate schema version
        schema_version = data.get("schema_version", SUPPORTED_SCHEMA_VERSION)
        if schema_version != SUPPORTED_SCHEMA_VERSION:
            raise ValueError(
                f"Unsupported schema version: {schema_version}. "
                f"Expected: {SUPPORTED_SCHEMA_VERSION}"
            )

        # Extract submesh types data
        submesh_types_data = data.get("submesh_types")
        if submesh_types_data is None:
            raise KeyError("Missing 'submesh_types' section in JSON data.")

        # Reconstruct submesh types
        reconstructed_submesh_types = {}
        for domain, submesh_info in submesh_types_data.items():
            try:
                module_name = submesh_info["module"]
                class_name = submesh_info["class"]

                # Import module and get class
                module = importlib.import_module(module_name)
                submesh_class = getattr(module, class_name)

                # Wrap in MeshGenerator to match the expected format
                reconstructed_submesh_types[domain] = pybamm.MeshGenerator(
                    submesh_class
                )

            except (ModuleNotFoundError, AttributeError) as e:
                raise ImportError(
                    f"Could not import submesh type '{class_name}' from '{module_name}': {e}"
                ) from e
            except Exception as e:
                raise ValueError(
                    f"Failed to reconstruct submesh type for domain '{domain}': {e}"
                ) from e

        return reconstructed_submesh_types

    @staticmethod
    def _create_symbol_key(symbol_json: dict) -> str:
        """
        Given the JSON‐dict for a symbol, return a unique, hashable key.
        We just sort the dict keys and dump to a string.
        """
        return json.dumps(symbol_json, sort_keys=True)

    @staticmethod
    def load_custom_model(filename: str | dict) -> pybamm.BaseModel:
        """
        Loads a custom (symbolic) PyBaMM model from a JSON file or dictionary.

        Reconstructs a model saved using `save_custom_model`, including its rhs,
        algebraic equations, initial and boundary conditions, events, and variables.
        Returns a fully symbolic model ready for further processing or discretisation.

        Parameters
        ----------
        filename : str or dict
            Path to the JSON file containing the saved model, or a dictionary
            containing the serialised model data.

        Returns
        -------
        :class:`pybamm.BaseModel` or subclass
            The reconstructed symbolic PyBaMM model.

        Example
        -------
        >>> import pybamm
        >>> model = pybamm.lithium_ion.BasicDFN()
        >>> from pybamm.expression_tree.operations.serialise import Serialise
        >>> Serialise.save_custom_model(model, "basicdfn_model.json")
        >>> loaded_model = Serialise.load_custom_model("basicdfn_model.json")

        """
        if isinstance(filename, dict):
            data = filename
        else:
            try:
                with open(filename) as file:
                    data = json.load(file)
            except FileNotFoundError as err:
                raise FileNotFoundError(f"Could not find file: {filename}") from err
            except json.JSONDecodeError as e:
                raise pybamm.InvalidModelJSONError(
                    f"The model defined in the file '{filename}' contains invalid JSON: {e!s}"
                ) from e

        # Validate outer structure
        schema_version = data.get("schema_version", SUPPORTED_SCHEMA_VERSION)
        if schema_version != SUPPORTED_SCHEMA_VERSION:
            raise ValueError(
                f"Unsupported schema version: {schema_version}. "
                f"Expected: {SUPPORTED_SCHEMA_VERSION}"
            )

        model_data = data.get("model")
        if model_data is None:
            raise KeyError("Missing 'model' section in JSON file.")

        required = [
            "name",
            "rhs",
            "initial_conditions",
            "base_class",
            "algebraic",
            "boundary_conditions",
            "events",
            "variables",
        ]
        missing = [k for k in required if k not in model_data]
        if missing:
            raise KeyError(f"Missing required model sections: {missing}")

        battery_model = model_data.get("base_class")
        if not battery_model or battery_model.strip() == "pybamm.BaseModel":
            base_cls = pybamm.BaseModel
        else:
            module_name, class_name = battery_model.rsplit(".", 1)
            try:
                module = importlib.import_module(module_name)
                base_cls = getattr(module, class_name)
            except (ModuleNotFoundError, AttributeError) as e:
                if battery_model == "builtins.object":
                    base_cls = pybamm.BaseModel
                else:
                    raise ImportError(
                        f"Could not import base class '{battery_model}': {e}"
                    ) from e

        model = base_cls()
        model.name = model_data["name"]
        model.schema_version = schema_version

        all_variable_keys = (
            [lhs_json for lhs_json, _ in model_data["rhs"]]
            + [lhs_json for lhs_json, _ in model_data["initial_conditions"]]
            + [lhs_json for lhs_json, _ in model_data["algebraic"]]
            + [variable_json for variable_json, _ in model_data["boundary_conditions"]]
        )

        symbol_map = {}
        for variable_json in all_variable_keys:
            try:
                symbol = convert_symbol_from_json(variable_json)
                key = Serialise._create_symbol_key(variable_json)
                symbol_map[key] = symbol
            except Exception as e:
                raise ValueError(
                    f"Failed to process symbol key for variable {variable_json}: {e!s}"
                ) from e

        model.rhs = {}
        for lhs_json, rhs_expr_json in model_data["rhs"]:
            try:
                lhs = symbol_map[Serialise._create_symbol_key(lhs_json)]
                rhs = convert_symbol_from_json(rhs_expr_json)
                model.rhs[lhs] = rhs
            except Exception as e:
                raise ValueError(
                    f"Failed to convert rhs entry for {lhs_json}: {e!s}"
                ) from e

        model.algebraic = {}
        for lhs_json, algebraic_expr_json in model_data["algebraic"]:
            try:
                lhs = symbol_map[Serialise._create_symbol_key(lhs_json)]
                rhs = convert_symbol_from_json(algebraic_expr_json)
                model.algebraic[lhs] = rhs
            except Exception as e:
                raise ValueError(
                    f"Failed to convert algebraic entry for {lhs_json}: {e!s}"
                ) from e

        model.initial_conditions = {}
        for lhs_json, initial_value_json in model_data["initial_conditions"]:
            try:
                lhs = symbol_map[Serialise._create_symbol_key(lhs_json)]
                rhs = convert_symbol_from_json(initial_value_json)
                model.initial_conditions[lhs] = rhs
            except Exception as e:
                raise ValueError(
                    f"Failed to convert initial condition entry for {lhs_json}: {e!s}"
                ) from e

        model.boundary_conditions = {}
        for variable_json, condition_dict in model_data["boundary_conditions"]:
            try:
                variable = symbol_map[Serialise._create_symbol_key(variable_json)]
                sides = {}
                for side, (expression_json, boundary_type) in condition_dict.items():
                    try:
                        expr = convert_symbol_from_json(expression_json)
                        sides[side] = (expr, boundary_type)
                    except Exception as e:
                        raise ValueError(
                            f"Failed to convert boundary expression for variable {variable_json} on side '{side}': {e!s}"
                        ) from e
                model.boundary_conditions[variable] = sides
            except Exception as e:
                raise ValueError(
                    f"Failed to convert boundary condition entry for variable {variable_json}: {e!s}"
                ) from e

        model.events = []
        for event_data in model_data["events"]:
            try:
                name = event_data["name"]
                expr = convert_symbol_from_json(event_data["expression"])
                event_type = event_data["event_type"]
                model.events.append(pybamm.Event(name, expr, event_type))
            except Exception as e:
                raise ValueError(
                    f"Failed to convert event '{event_data.get('name', 'UNKNOWN')}': {e!s}"
                ) from e

        model.variables = {}
        for variable_name, expression_json in model_data["variables"].items():
            try:
                key = Serialise._create_symbol_key(expression_json)
                symbol = symbol_map.get(key)
                if symbol is None:
                    symbol = convert_symbol_from_json(expression_json)
                model.variables[variable_name] = symbol
            except Exception as e:
                raise ValueError(
                    f"Failed to convert variable '{variable_name}': {e!s}"
                ) from e

        return model

    @staticmethod
    def save_parameters(parameters: dict, filename=None):
        """
        Serializes a dictionary of parameters to a JSON file.
        The values can be numbers, PyBaMM symbols, or callables.

        Parameters
        ----------
        parameters : dict
            A dictionary of parameter names and values.
            Values can be numeric, PyBaMM symbols, or callables.

        filename : str, optional
            If given, saves the serialized parameters to this file.
        """
        parameter_values_dict = {}

        for k, v in parameters.items():
            if callable(v):
                parameter_values_dict[k] = convert_symbol_to_json(
                    convert_function_to_symbolic_expression(v, k)
                )
            else:
                parameter_values_dict[k] = convert_symbol_to_json(v)

        if filename is not None:
            with open(filename, "w") as f:
                json.dump(parameter_values_dict, f, indent=4)

    @staticmethod
    def load_parameters(filename):
        """
        Load a JSON file of parameters (either from Serialise.save_parameters
        or from a standard pybamm.ParameterValues.save), and return a
        pybamm.ParameterValues object.

        - If a value is a dict with a "type" key, deserialize it as a PyBaMM symbol.
        - Otherwise (float, int, bool, str, list, dict-without-type), leave it as-is.
        """
        with open(filename) as f:
            raw_dict = json.load(f)

        deserialized = {}
        for key, val in raw_dict.items():
            if isinstance(val, dict) and "type" in val:
                deserialized[key] = convert_symbol_from_json(val)

            elif isinstance(val, list):
                deserialized[key] = val

            elif isinstance(val, (numbers.Number | bool)):
                deserialized[key] = val

            elif isinstance(val, str):
                deserialized[key] = val

            elif isinstance(val, dict):
                deserialized[key] = val

            else:
                raise ValueError(
                    f"Unsupported parameter format for key '{key}': {val!r}"
                )

        return pybamm.ParameterValues(deserialized)

    # Helper functions

    def _get_pybamm_class(self, snippet: dict):
        """Find a pybamm class to initialise from object path"""
        parts = snippet["py/object"].split(".")
        module = importlib.import_module(".".join(parts[:-1]))

        class_ = getattr(module, parts[-1])

        try:
            empty_class = self._Empty()
            empty_class.__class__ = class_

            return empty_class

        except TypeError:
            # Mesh objects have a different layouts
            empty_dict_class = self._EmptyDict()
            empty_dict_class.__class__ = class_

            return empty_dict_class

    def _deconstruct_pybamm_dicts(self, dct: dict):
        """
        Converts dictionaries which contain pybamm classes as keys
        into a json serialisable format.

        Dictionary keys present as pybamm objects are given a seperate key
        as "symbol_<symbol name>" to store the dictionary required to reconstruct
        a symbol, and their seperate key is used in the original dictionary. E.G:

        {'rod':
            {SpatialVariable(name='spat_var'): {"min":0.0, "max":2.0} }
            }
        converts to

        {'rod':
            {'symbol_spat_var': {"min":0.0, "max":2.0} },
        'spat_var':
            {"py/object":pybamm....}
        }

        Dictionaries which don't contain pybamm symbols are returned unchanged.
        """

        def nested_convert(obj):
            if isinstance(obj, dict):
                new_dict = {}
                for k, v in obj.items():
                    if isinstance(k, pybamm.Symbol):
                        new_k = self._SymbolEncoder().default(k)
                        new_dict["symbol_" + new_k["name"]] = new_k
                        k = new_k["name"]
                    new_dict[k] = nested_convert(v)
                return new_dict
            return obj

        try:
            _ = json.dumps(dct)
            return dict(dct)
        except TypeError:  # dct must contain pybamm objects
            return nested_convert(dct)

    def _reconstruct_symbol(self, dct: dict):
        """Reconstruct an individual pybamm Symbol"""
        symbol_class = self._get_pybamm_class(dct)
        symbol = symbol_class._from_json(dct)
        return symbol

    def _reconstruct_expression_tree(self, node: dict):
        """
        Loop through an expression tree creating pybamm Symbol classes

        Conducts post-order tree traversal to turn each tree node into a
        `pybamm.Symbol` class, starting from leaf nodes without children and
        working upwards.

        Parameters
        ----------
        node: dict
            A node in an expression tree.
        """
        if "children" in node:
            for i, c in enumerate(node["children"]):
                child_obj = self._reconstruct_expression_tree(c)
                node["children"][i] = child_obj
        elif "expression" in node:
            expression_obj = self._reconstruct_expression_tree(node["expression"])
            node["expression"] = expression_obj

        obj = self._reconstruct_symbol(node)

        return obj

    def _reconstruct_mesh(self, node: dict):
        """Reconstructs a Mesh object"""
        if "sub_meshes" in node:
            for k, v in node["sub_meshes"].items():
                sub_mesh = self._reconstruct_symbol(v)
                node["sub_meshes"][k] = sub_mesh

        new_mesh = self._reconstruct_symbol(node)

        return new_mesh

    def _reconstruct_pybamm_dict(self, obj: dict):
        """
        pybamm.Geometry can contain PyBaMM symbols as dictionary keys.

        Converts
        {"rod":
            {"symbol_spat_var":
                {"min":0.0, "max":2.0} },
            "spat_var":
                {"py/object":"pybamm...."}
        }

        from an exported JSON file to

        {"rod":
            {SpatialVariable(name="spat_var"): {"min":0.0, "max":2.0} }
            }

        """

        def recurse(obj):
            if isinstance(obj, dict):
                new_dict = {}
                for k, v in obj.items():
                    if "symbol_" in k:
                        new_dict[k] = self._reconstruct_symbol(v)
                    elif isinstance(v, dict):
                        new_dict[k] = recurse(v)
                    else:
                        new_dict[k] = v

                pattern = re.compile("symbol_")
                symbol_keys = {k: v for k, v in new_dict.items() if pattern.match(k)}

                # rearrange the dictionary to make pybamm objects the dictionary keys
                if symbol_keys:
                    for k, v in symbol_keys.items():
                        new_dict[v] = new_dict[k.lstrip("symbol_")]
                        del new_dict[k]
                        del new_dict[k.lstrip("symbol_")]

                return new_dict
            return obj

        return recurse(obj)

    def _convert_options(self, d):
        """
        Converts a dictionary with nested lists to nested tuples,
        used to convert model options back into correct format
        """
        if isinstance(d, dict):
            return {k: self._convert_options(v) for k, v in d.items()}
        elif isinstance(d, list):
            return tuple(self._convert_options(item) for item in d)
        else:
            return d


def convert_function_to_symbolic_expression(func, name=None):
    """
    Converts a Python function to a PyBaMM symbolic expression

    Parameters
    ----------
    func : callable
        The Python function to convert

    name : str, optional
        The name of the function to use in the symbolic expression. If not provided,
        the name of the function is used.

    Returns
    -------
    pybamm.Symbol
        The PyBaMM symbolic expression
    """
    # Create symbolic parameters for each input argument
    try:
        func_name = func.get_name()
        func_args = func.get_args()
        # Use the underlying function for evaluation
        func_to_eval = func.func
    except AttributeError:
        try:
            func_name = func.__name__
            func_args = list(inspect.signature(func).parameters)
            func_to_eval = func
        except AttributeError:
            # One more fallback, in case it's a partial
            func_name = func.func.__name__
            func_args = list(inspect.signature(func).parameters)
            func_to_eval = func

    sym_inputs = [pybamm.Parameter(arg) for arg in func_args]

    # Evaluate the function with symbolic inputs to get symbolic expression
    sym_output = func_to_eval(*sym_inputs)

    # Wrap the symbolic expression in an ExpressionFunctionParameter to allow access
    # to the function name and arguments
    name = name or func_name
    return ExpressionFunctionParameter(name, sym_output, func_name, func_args)


def convert_symbol_from_json(json_data):
    """
    Recursively converts a JSON dictionary back into PyBaMM symbolic expressions

    Parameters
    ----------
    json_data : dict
        Dictionary containing the serialized PyBaMM expression

    Returns
    -------
    pybamm.Symbol
        The reconstructed PyBaMM symbolic expression
    """
    if isinstance(json_data, float | int | bool):
        return json_data

    if isinstance(json_data, str):
        raise ValueError(f"Unexpected raw string in JSON: {json_data}")

    if json_data is None:
        return None
    if "type" not in json_data:
        raise ValueError(f"Missing 'type' key in JSON data: {json_data}")
    if isinstance(json_data, numbers.Number | list):
        return json_data
    elif json_data["type"] == "Parameter":
        # Convert stored parameters back to PyBaMM Parameter objects
        return pybamm.Parameter(json_data["name"])
    elif json_data["type"] == "InputParameter":
        return pybamm.InputParameter(json_data["name"])
    elif json_data["type"] == "Scalar":
        # Convert stored numerical values back to PyBaMM Scalar objects
        return pybamm.Scalar(json_data["value"])
    elif json_data["type"] == "Interpolant":
        return pybamm.Interpolant(
            [np.array(x) for x in json_data["x"]],
            np.array(json_data["y"]),
            [convert_symbol_from_json(c) for c in json_data["children"]],
            name=json_data["name"],
            interpolator=json_data["interpolator"],
            entries_string=json_data["entries_string"],
        )
    elif json_data["type"] == "FunctionParameter":
        diff_variable = json_data["diff_variable"]
        if diff_variable is not None:
            diff_variable = convert_symbol_from_json(diff_variable)
        # Use the parameter name as print_name to avoid showing
        # 'convert_symbol_from_json' in displays
        return pybamm.FunctionParameter(
            json_data["name"],
            {k: convert_symbol_from_json(v) for k, v in json_data["inputs"].items()},
            diff_variable=diff_variable,
            print_name=json_data["name"],
        )
    elif json_data["type"] == "ExpressionFunctionParameter":
        return ExpressionFunctionParameter(
            json_data["name"],
            convert_symbol_from_json(json_data["children"][0]),
            json_data["func_name"],
            json_data["func_args"],
        )
    elif json_data["type"] == "PrimaryBroadcast":
        child = convert_symbol_from_json(json_data["children"][0])
        domain = json_data["broadcast_domain"]
        return pybamm.PrimaryBroadcast(child, domain)
    elif json_data["type"] == "FullBroadcast":
        child = convert_symbol_from_json(json_data["children"][0])
        domains = json_data["domains"]
        return pybamm.FullBroadcast(child, broadcast_domains=domains)
    elif json_data["type"] == "SecondaryBroadcast":
        child = convert_symbol_from_json(json_data["children"][0])
        domain = json_data["broadcast_domain"]
        return pybamm.SecondaryBroadcast(child, domain)
    elif json_data["type"] == "BoundaryValue":
        child = convert_symbol_from_json(json_data["children"][0])
        side = json_data["side"]
        return pybamm.BoundaryValue(child, side)
    elif json_data["type"] == "Variable":
        bounds = tuple(
            convert_symbol_from_json(b)
            for b in json_data.get("bounds", [-float("inf"), float("inf")])
        )
        return pybamm.Variable(
            json_data["name"],
            domains=json_data["domains"],
            bounds=bounds,
        )
    elif json_data["type"] == "IndefiniteIntegral":
        child = convert_symbol_from_json(json_data["children"][0])
        integration_var_json = json_data["integration_variable"]
        integration_variable = convert_symbol_from_json(integration_var_json)
        if not isinstance(integration_variable, pybamm.SpatialVariable):
            raise TypeError(
                f"Expected SpatialVariable, got {type(integration_variable)}"
            )
        return pybamm.IndefiniteIntegral(child, [integration_variable])
    elif json_data["type"] == "SpatialVariable":
        return pybamm.SpatialVariable(
            json_data["name"],
            coord_sys=json_data.get("coord_sys", "cartesian"),
            domains=json_data.get("domains"),
        )
    elif json_data["type"] == "Time":
        return pybamm.Time()
    elif json_data["type"] == "Symbol":
        return pybamm.Symbol(
            json_data["name"],
            domains=json_data.get("domains", {}),
        )
    elif "children" in json_data:
        return getattr(pybamm, json_data["type"])(
            *[convert_symbol_from_json(c) for c in json_data["children"]]
        )
    else:
        raise ValueError(f"Unknown symbol type: {json_data['type']}")


def convert_symbol_to_json(symbol):
    """
    Converts a PyBaMM symbolic expression to a JSON-serializable dictionary

    Parameters
    ----------
    symbol : pybamm.Symbol
        The PyBaMM symbolic expression to convert

    Returns
    -------
    dict
        The JSON-serializable dictionary
    """
    if isinstance(symbol, ExpressionFunctionParameter):
        return {
            "type": "ExpressionFunctionParameter",
            "name": symbol.name,
            "children": [convert_symbol_to_json(symbol.child)],
            "func_name": symbol.func_name,
            "func_args": symbol.func_args,
        }
    elif isinstance(symbol, numbers.Number | list):
        return symbol
    elif isinstance(symbol, pybamm.Parameter):
        # Parameters are stored with their type and name
        return {"type": "Parameter", "name": symbol.name}
    elif isinstance(symbol, pybamm.Scalar):
        # Scalar values are stored with their numerical value
        return {"type": "Scalar", "value": symbol.value}
    elif isinstance(symbol, pybamm.SpecificFunction):
        if symbol.__class__ == pybamm.SpecificFunction:
            raise NotImplementedError("SpecificFunction is not supported directly")
        else:
            # Subclasses of SpecificFunction (e.g. Exp, Sin, etc.) can be reconstructed
            # from only the children
            return {
                "type": symbol.__class__.__name__,
                "children": [convert_symbol_to_json(c) for c in symbol.children],
            }
    elif isinstance(symbol, pybamm.PrimaryBroadcast):
        json_dict = {
            "type": "PrimaryBroadcast",
            "children": [convert_symbol_to_json(symbol.child)],
            "broadcast_domain": symbol.broadcast_domain,
        }
        return json_dict
    elif isinstance(symbol, pybamm.IndefiniteIntegral):
        integration_var = (
            symbol.integration_variable[0]
            if isinstance(symbol.integration_variable, list)
            else symbol.integration_variable
        )
        json_dict = {
            "type": "IndefiniteIntegral",
            "children": [convert_symbol_to_json(symbol.child)],
            "integration_variable": convert_symbol_to_json(integration_var),
        }
        return json_dict
    elif isinstance(symbol, pybamm.BoundaryValue):
        json_dict = {
            "type": "BoundaryValue",
            "side": symbol.side,
            "children": [convert_symbol_to_json(symbol.orphans[0])],
        }
        return json_dict
    elif isinstance(symbol, pybamm.SecondaryBroadcast):
        json_dict = {
            "type": "SecondaryBroadcast",
            "children": [convert_symbol_to_json(symbol.child)],
            "broadcast_domain": symbol.broadcast_domain,
        }
        return json_dict
    elif isinstance(symbol, pybamm.FullBroadcast):
        json_dict = {
            "type": "FullBroadcast",
            "children": [convert_symbol_to_json(symbol.child)],
            "domains": symbol.domains,
        }
        return json_dict
    elif isinstance(symbol, pybamm.Interpolant):
        return {
            "type": symbol.__class__.__name__,
            "x": [x.tolist() for x in symbol.x],
            "y": symbol.y.tolist(),
            "children": [convert_symbol_to_json(c) for c in symbol.children],
            "name": symbol.name,
            "interpolator": symbol.interpolator,
            "entries_string": symbol.entries_string,
        }
    elif isinstance(symbol, pybamm.Variable):
        json_dict = {
            "type": "Variable",
            "name": symbol.name,
            "domains": symbol.domains,
            "bounds": [
                convert_symbol_to_json(symbol.bounds[0]),
                convert_symbol_to_json(symbol.bounds[1]),
            ],
        }
        return json_dict
    elif isinstance(symbol, pybamm.ConcatenationVariable):
        json_dict = {
            "type": "ConcatenationVariable",
            "name": symbol.name,
            "children": [convert_symbol_to_json(child) for child in symbol.children],
        }
        return json_dict
    elif isinstance(symbol, pybamm.Time):
        return {"type": "Time"}
    elif isinstance(symbol, pybamm.FunctionParameter):
        input_names = symbol.input_names
        inputs = {
            input_names[i]: convert_symbol_to_json(symbol.orphans[i])
            for i in range(len(input_names))
        }
        diff_variable = symbol.diff_variable
        if diff_variable is not None:
            diff_variable = convert_symbol_to_json(diff_variable)
        return {
            "type": symbol.__class__.__name__,
            "inputs": inputs,
            "diff_variable": diff_variable,
            "name": symbol.name,
        }
    elif isinstance(symbol, pybamm.Symbol):
        # Generic fallback for other symbols with children
        json_dict = {
            "type": symbol.__class__.__name__,
            "domains": symbol.domains,
            "children": [convert_symbol_to_json(c) for c in symbol.children],
        }
        if hasattr(symbol, "name"):
            json_dict["name"] = symbol.name
        return json_dict
    else:
        raise ValueError(
            f"Error processing '{symbol.name}'. Unknown symbol type: {type(symbol)}"
        )



#
# BatchStudy class
#
from itertools import product

import pybamm


class BatchStudy:
    """
    A BatchStudy class for comparison of different PyBaMM simulations.

    Parameters
    ----------
    models : dict
        A dictionary of models to be simulated
    experiments : dict (optional)
        A dictionary of experimental conditions under which to solve the model.
        Default is None
    geometries : dict (optional)
        A dictionary of geometries upon which to solve the model
    parameter_values : dict (optional)
        A dictionary of parameters and their corresponding numerical values.
        Default is None
    submesh_types : dict (optional)
        A dictionary of the types of submesh to use on each subdomain.
        Default is None
    var_pts : dict (optional)
        A dictionary of the number of points used by each spatial variable.
        Default is None
    spatial_methods : dict (optional)
        A dictionary of the types of spatial method to use on each domain.
        Default is None
    solvers : dict (optional)
        A dictionary of solvers to use to solve the model. Default is None
    output_variables : dict (optional)
        A dictionary of variables to plot automatically. Default is None
    C_rates : dict (optional)
        A dictionary of C-rates at which you would like to run a constant current
        (dis)charge. Default is None
    repeats : int (optional)
        The number of times `solve` should be called. Default is 1
    permutations : bool (optional)
        If False runs first model with first solver, first experiment
        and second model with second solver, second experiment etc.
        If True runs a cartesian product of models, solvers and experiments.
        Default is False
    """

    INPUT_LIST = [
        "experiments",
        "geometries",
        "parameter_values",
        "submesh_types",
        "var_pts",
        "spatial_methods",
        "solvers",
        "output_variables",
        "C_rates",
    ]

    def __init__(
        self,
        models,
        experiments=None,
        geometries=None,
        parameter_values=None,
        submesh_types=None,
        var_pts=None,
        spatial_methods=None,
        solvers=None,
        output_variables=None,
        C_rates=None,
        repeats=1,
        permutations=False,
    ):
        self.models = models
        self.experiments = experiments
        self.geometries = geometries
        self.parameter_values = parameter_values
        self.submesh_types = submesh_types
        self.var_pts = var_pts
        self.spatial_methods = spatial_methods
        self.solvers = solvers
        self.output_variables = output_variables
        self.C_rates = C_rates
        self.repeats = repeats
        self.permutations = permutations
        self.quick_plot = None

        if not self.permutations:
            for name in self.INPUT_LIST:
                if getattr(self, name) and (
                    len(self.models) != len(getattr(self, name))
                ):
                    raise ValueError(
                        f"Either provide no {name} or an equal number of {name}"
                        f" as the models ({len(self.models)} models given)"
                        f" if permutations=False"
                    )

    def solve(
        self,
        t_eval=None,
        solver=None,
        save_at_cycles=None,
        calc_esoh=True,
        starting_solution=None,
        initial_soc=None,
        t_interp=None,
        **kwargs,
    ):
        """
        For more information on the parameters used in the solve,
        See :meth:`pybamm.Simulation.solve`
        """
        self.sims = []
        iter_func = product if self.permutations else zip

        # Instantiate items in INPUT_LIST based on the value of self.permutations
        inp_values = []
        for name in self.INPUT_LIST:
            if getattr(self, name):
                inp_value = getattr(self, name).values()
            elif self.permutations:
                inp_value = [None]
            else:
                inp_value = [None] * len(self.models)
            inp_values.append(inp_value)

        for (
            model,
            experiment,
            geometry,
            parameter_value,
            submesh_type,
            var_pt,
            spatial_method,
            solver,
            output_variable,
            C_rate,
        ) in iter_func(self.models.values(), *inp_values):
            sim = pybamm.Simulation(
                model,
                experiment=experiment,
                geometry=geometry,
                parameter_values=parameter_value,
                submesh_types=submesh_type,
                var_pts=var_pt,
                spatial_methods=spatial_method,
                solver=solver,
                output_variables=output_variable,
                C_rate=C_rate,
            )
            # Repeat to get average solve time and integration time
            solve_time = 0
            integration_time = 0
            for _ in range(self.repeats):
                sol = sim.solve(
                    t_eval,
                    solver,
                    save_at_cycles,
                    calc_esoh,
                    starting_solution,
                    initial_soc,
                    **kwargs,
                )
                solve_time += sol.solve_time
                integration_time += sol.integration_time
            sim.solution.solve_time = solve_time / self.repeats
            sim.solution.integration_time = integration_time / self.repeats
            self.sims.append(sim)

    def plot(self, output_variables=None, **kwargs):
        """
        For more information on the parameters used in the plot,
        See :meth:`pybamm.Simulation.plot`
        """
        self.quick_plot = pybamm.dynamic_plot(
            self.sims, output_variables=output_variables, **kwargs
        )
        return self.quick_plot

    def create_gif(self, number_of_images=80, duration=0.1, output_filename="plot.gif"):
        """
        Generates x plots over a time span of t_eval and compiles them to create
        a GIF. For more information see :meth:`pybamm.QuickPlot.create_gif`

        Parameters
        ----------
        number_of_images : int, optional
            Number of images/plots to be compiled for a GIF.
        duration : float, optional
            Duration of visibility of a single image/plot in the created GIF.
        output_filename : str, optional
            Name of the generated GIF file.

        """
        if not hasattr(self, "sims"):
            raise ValueError("The simulations have not been solved yet.")
        if self.quick_plot is None:
            self.quick_plot = pybamm.QuickPlot(self.sims)

        self.quick_plot.create_gif(
            number_of_images=number_of_images,
            duration=duration,
            output_filename=output_filename,
        )



#
# No convection
#
import pybamm

from .base_through_cell_convection import BaseThroughCellModel


class NoConvection(BaseThroughCellModel):
    """A submodel for case where there is no convection.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param, options=options)

    def get_fundamental_variables(self):
        variables = {}
        for domain in self.options.whole_cell_domains:
            if domain != "separator":
                v_box_k = pybamm.FullBroadcast(0, domain, "current collector")
                div_v_box_k = pybamm.FullBroadcast(0, domain, "current collector")
                p_k = pybamm.FullBroadcast(0, domain, "current collector")

                variables.update(
                    self._get_standard_convection_variables(
                        domain, v_box_k, div_v_box_k, p_k
                    )
                )

        return variables

    def get_coupled_variables(self, variables):
        # Simple formula for velocity in the separator
        v_box_s = pybamm.FullBroadcast(0, "separator", "current collector")
        div_v_box_s = pybamm.FullBroadcast(0, "separator", "current collector")

        variables.update(
            self._get_standard_sep_velocity_variables(v_box_s, div_v_box_s)
        )
        variables.update(self._get_standard_whole_cell_velocity_variables(variables))
        variables.update(
            self._get_standard_whole_cell_acceleration_variables(variables)
        )
        variables.update(self._get_standard_whole_cell_pressure_variables(variables))

        return variables



#
# Class for when there is no oxygen
#
import pybamm

from .base_oxygen_diffusion import BaseModel


class NoOxygen(BaseModel):
    """Class for when there is no oxygen

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    """

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):
        c_ox_n = pybamm.FullBroadcast(0, ["negative electrode"], "current collector")
        c_ox_s = pybamm.FullBroadcast(0, ["separator"], "current collector")
        c_ox_p = pybamm.FullBroadcast(0, ["positive electrode"], "current collector")

        variables = self._get_standard_concentration_variables(c_ox_n, c_ox_s, c_ox_p)

        N_e = pybamm.FullBroadcastToEdges(
            0,
            ["negative electrode", "separator", "positive electrode"],
            "current collector",
        )

        variables.update(self._get_standard_flux_variables(N_e))

        return variables



import pybamm


class ThermalSubModel(pybamm.BaseSubModel):
    """
    Thermal SubModel for use with equivalent
    circuits.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param)
        self.model_options = options

    def get_fundamental_variables(self):
        T_cell = pybamm.Variable("Cell temperature [degC]")
        T_jig = pybamm.Variable("Jig temperature [degC]")

        # Note this is defined in deg C
        T_amb = self.param.T_amb(pybamm.t)

        Q_cell_cool = -self.param.k_cell_jig * (T_cell - T_jig)
        Q_jig_cool = -self.param.k_jig_air * (T_jig - T_amb)

        kelvin = 273.15
        variables = {
            "Cell temperature [degC]": T_cell,
            "Cell temperature [K]": T_cell + kelvin,
            "Jig temperature [degC]": T_jig,
            "Jig temperature [K]": T_jig + kelvin,
            "Ambient temperature [degC]": T_amb,
            "Ambient temperature [K]": T_amb + kelvin,
            "Heat transfer from cell to jig [W]": Q_cell_cool,
            "Heat transfer from jig to ambient [W]": Q_jig_cool,
        }

        return variables

    def get_coupled_variables(self, variables):
        number_of_rc_elements = self.model_options["number of rc elements"]
        number_of_elements = number_of_rc_elements + 1

        Q_irr = pybamm.Scalar(0)
        for i in range(number_of_elements):
            Q_irr += variables[f"Element-{i} irreversible heat generation [W]"]

        Q_rev = variables["Reversible heat generation [W]"]

        variables.update(
            {
                "Irreversible heat generation [W]": Q_irr,
                "Total heat generation [W]": Q_irr + Q_rev,
            }
        )

        return variables

    def set_rhs(self, variables):
        T_cell = variables["Cell temperature [degC]"]
        T_jig = variables["Jig temperature [degC]"]

        Q_irr = variables["Irreversible heat generation [W]"]
        Q_rev = variables["Reversible heat generation [W]"]

        Q_cell_cool = variables["Heat transfer from cell to jig [W]"]
        Q_jig_cool = variables["Heat transfer from jig to ambient [W]"]

        self.rhs = {
            T_cell: (Q_irr + Q_rev + Q_cell_cool) / self.param.cth_cell,
            T_jig: (Q_jig_cool - Q_cell_cool) / self.param.cth_jig,
        }

    def set_initial_conditions(self, variables):
        T_cell = variables["Cell temperature [degC]"]
        T_jig = variables["Jig temperature [degC]"]
        self.initial_conditions = {
            T_cell: self.param.initial_T_cell,
            T_jig: self.param.initial_T_jig,
        }



#
# Processed Variable Computed class
#
from __future__ import annotations

import casadi
import numpy as np
import xarray as xr
from scipy.integrate import cumulative_trapezoid

import pybamm

from .base_processed_variable import BaseProcessedVariable


class ProcessedVariableComputed(BaseProcessedVariable):
    """
    An object that can be evaluated at arbitrary (scalars or vectors) t and x, and
    returns the (interpolated) value of the base variable at that t and x.

    The 'Computed' variant of ProcessedVariable deals with variables that have
    been derived at solve time (see the 'output_variables' solver option),
    where the full state-vector is not itself propagated and returned.

    Parameters
    ----------
    base_variables : list of :class:`pybamm.Symbol`
        A list of base variables with a method `evaluate(t,y)`, each entry of which
        returns the value of that variable for that particular sub-solution.
        A Solution can be comprised of sub-solutions which are the solutions of
        different models.
        Note that this can be any kind of node in the expression tree, not
        just a :class:`pybamm.Variable`.
        When evaluated, returns an array of size (m,n)
    base_variables_casadi : list of :class:`casadi.Function`
        A list of casadi functions. When evaluated, returns the same thing as
        `base_Variable.evaluate` (but more efficiently).
    base_variable_data : list of :numpy:array
        A list of numpy arrays, the returned evaluations.
    solution : :class:`pybamm.Solution`
        The solution object to be used to create the processed variables
    warn : bool, optional
        Whether to raise warnings when trying to evaluate time and length scales.
        Default is True.
    time_indep : bool, optional
        Whether the variable is time-independent. Default is False. Used for
        time integral or sum variables
    """

    def __init__(
        self,
        base_variables,
        base_variables_casadi,
        base_variables_data,
        solution,
        cumtrapz_ic=None,
        time_indep=False,
    ):
        self.base_variables = base_variables
        self.base_variables_casadi = base_variables_casadi
        self.base_variables_data = base_variables_data

        self.all_ts = solution.all_ts
        self.all_ys = solution.all_ys
        self.all_inputs = solution.all_inputs
        self.all_inputs_casadi = solution.all_inputs_casadi

        self.mesh = base_variables[0].mesh
        self.domain = base_variables[0].domain
        self.domains = base_variables[0].domains
        self.cumtrapz_ic = cumtrapz_ic
        self.time_indep = time_indep

        # Sensitivity starts off uninitialized, only set when called
        self._sensitivities = None
        self.solution_sensitivities = solution.sensitivities

        # Store time
        self.t_pts = solution.t

        # Evaluate base variable at initial time
        self.base_eval_shape = self.base_variables[0].shape
        self.base_eval_size = self.base_variables[0].size
        self.unroll_params = {}

        # handle 2D or 3D (in space) finite element variables differently
        if (
            self.mesh
            and "current collector" in self.domain
            and isinstance(self.mesh, pybamm.ScikitSubMesh2D)
        ):
            self.initialise_2D_scikit_fem()
            return
        if hasattr(base_variables[0], "secondary_mesh"):
            if "current collector" in base_variables[0].domains[
                "secondary"
            ] and isinstance(base_variables[0].secondary_mesh, pybamm.ScikitSubMesh2D):
                self.initialise_3D_scikit_fem()
                return

        # check variable shape
        if len(self.base_eval_shape) == 0 or self.base_eval_shape[0] == 1:
            if self.time_indep:
                self.initialise_time_independent()
            else:
                self.initialise_0D()
            return

        n = self.mesh.npts
        base_shape = self.base_eval_shape[0]
        # Try some shapes that could make the variable a 1D variable
        if base_shape in [n, n + 1]:
            self.initialise_1D()
            return

        # Try some shapes that could make the variable a 2D variable
        first_dim_nodes = self.mesh.nodes
        first_dim_edges = self.mesh.edges
        second_dim_pts = self.base_variables[0].secondary_mesh.nodes
        if self.base_eval_size // len(second_dim_pts) in [
            len(first_dim_nodes),
            len(first_dim_edges),
        ]:
            self.initialise_2D()
            return

        # Try some shapes that could make the variable a 3D variable
        tertiary_pts = self.base_variables[0].tertiary_mesh.nodes
        if self.base_eval_size // (len(second_dim_pts) * len(tertiary_pts)) in [
            len(first_dim_nodes),
            len(first_dim_edges),
        ]:
            self.initialise_3D()
            return

        raise NotImplementedError(f"Shape not recognized for {base_variables[0]}")

    def as_computed(self) -> ProcessedVariableComputed:
        return self

    def add_sensitivity(self, param, data):
        # unroll from sparse representation into n-d matrix
        # then flatten for consistency with full state-vector
        # ProcessedVariable sensitivities
        self._sensitivities[param] = self.unroll(data).flatten()

    def _unroll_nnz(self, realdata=None):
        # unroll in nnz != numel, otherwise copy
        if realdata is None:
            realdata = self.base_variables_data
        if isinstance(self.base_variables_casadi[0], casadi.Function):  # casadi fcn
            sp = self.base_variables_casadi[0](0, 0, 0).sparsity()
            nnz = sp.nnz()
            numel = sp.numel()
            row = sp.row()
        if nnz != numel:
            data = [None] * len(realdata)
            for datak in range(len(realdata)):
                data[datak] = np.zeros(self.base_eval_shape[0] * len(self.t_pts))
                var_data = realdata[0].flatten()
                k = 0
                for t_i in range(len(self.t_pts)):
                    base = t_i * numel
                    for r in row:
                        data[datak][base + r] = var_data[k]
                        k = k + 1
        else:
            data = realdata
        return data

    def unroll_0D(self, realdata=None):
        if realdata is None:
            realdata = self.base_variables_data
        return np.concatenate(realdata, axis=0).flatten()

    def unroll_1D(self, realdata=None):
        len_space = self.base_eval_shape[0]
        return (
            np.concatenate(self._unroll_nnz(realdata), axis=0)
            .reshape((len(self.t_pts), len_space))
            .transpose()
        )

    def unroll_2D(self, realdata=None, n_dim1=None, n_dim2=None, axis_swaps=None):
        # initialise settings on first run
        if axis_swaps is None:
            axis_swaps = []
        if not self.unroll_params:
            self.unroll_params["n_dim1"] = n_dim1
            self.unroll_params["n_dim2"] = n_dim2
            self.unroll_params["axis_swaps"] = axis_swaps
        # use stored settings on subsequent runs
        if not n_dim1:
            n_dim1 = self.unroll_params["n_dim1"]
            n_dim2 = self.unroll_params["n_dim2"]
            axis_swaps = self.unroll_params["axis_swaps"]
        entries = np.concatenate(self._unroll_nnz(realdata), axis=0).reshape(
            (len(self.t_pts), n_dim1, n_dim2)
        )
        for a, b in axis_swaps:
            entries = np.moveaxis(entries, a, b)
        return entries

    def unroll_3D(
        self, realdata=None, n_dim1=None, n_dim2=None, n_dim3=None, axis_swaps=None
    ):
        if axis_swaps is None:
            axis_swaps = []
        if not self.unroll_params:
            self.unroll_params["n_dim1"] = n_dim1
            self.unroll_params["n_dim2"] = n_dim2
            self.unroll_params["n_dim3"] = n_dim3
            self.unroll_params["axis_swaps"] = axis_swaps
        if n_dim1 is None:
            n_dim1 = self.unroll_params["n_dim1"]
            n_dim2 = self.unroll_params["n_dim2"]
            n_dim3 = self.unroll_params["n_dim3"]
            axis_swaps = self.unroll_params["axis_swaps"]
        entries = (
            np.concatenate(self._unroll_nnz(realdata), axis=0)
            .transpose()
            .reshape(
                (len(self.t_pts), n_dim1, n_dim2, n_dim3),
            )
        )
        for a, b in axis_swaps:
            entries = np.moveaxis(entries, a, b)
        return entries

    def unroll(self, realdata=None):
        if self.dimensions == 0:
            return self.unroll_0D(realdata=realdata)
        elif self.dimensions == 1:
            return self.unroll_1D(realdata=realdata)
        elif self.dimensions == 2:
            return self.unroll_2D(realdata=realdata)
        elif self.dimensions == 3:
            return self.unroll_3D(realdata=realdata)
        else:
            raise NotImplementedError(f"Unsupported data dimension: {self.dimensions}")

    def initialise_time_independent(self):
        self.entries = self.unroll_0D()
        self._xr_data_array = None
        self.dimensions = 0

    def initialise_0D(self):
        entries = self.unroll_0D()

        if self.cumtrapz_ic is not None:
            entries = cumulative_trapezoid(
                entries, self.t_pts, initial=float(self.cumtrapz_ic)
            )

        # set up interpolation
        self._xr_data_array = xr.DataArray(entries, coords=[("t", self.t_pts)])

        self.entries = entries
        self.dimensions = 0

    def initialise_1D(self):
        entries = self.unroll_1D()

        # Get node and edge values
        nodes = self.mesh.nodes
        edges = self.mesh.edges
        if entries.shape[0] == len(nodes):
            space = nodes
        elif entries.shape[0] == len(edges):
            space = edges

        # add points outside domain for extrapolation to boundaries
        extrap_space_left = np.array([2 * space[0] - space[1]])
        extrap_space_right = np.array([2 * space[-1] - space[-2]])
        space = np.concatenate([extrap_space_left, space, extrap_space_right])
        extrap_entries_left = 2 * entries[0] - entries[1]
        extrap_entries_right = 2 * entries[-1] - entries[-2]
        entries_for_interp = np.vstack(
            [extrap_entries_left, entries, extrap_entries_right]
        )

        # assign attributes for reference (either x_sol or r_sol)
        self.entries = entries
        self.dimensions = 1
        if self.domain[0].endswith("particle"):
            self.first_dimension = "r"
            self.r_sol = space
        elif self.domain[0] in [
            "negative electrode",
            "separator",
            "positive electrode",
        ]:
            self.first_dimension = "x"
            self.x_sol = space
        elif self.domain == ["current collector"]:
            self.first_dimension = "z"
            self.z_sol = space
        elif self.domain[0].endswith("particle size"):
            self.first_dimension = "R"
            self.R_sol = space
        else:
            self.first_dimension = "x"
            self.x_sol = space

        # assign attributes for reference
        pts_for_interp = space
        self.internal_boundaries = self.mesh.internal_boundaries

        # Set first_dim_pts to edges for nicer plotting
        self.first_dim_pts = edges

        # set up interpolation
        self._xr_data_array = xr.DataArray(
            entries_for_interp,
            coords=[(self.first_dimension, pts_for_interp), ("t", self.t_pts)],
        )

    def initialise_2D(self):
        """
        Initialise a 2D object that depends on x and r, x and z, x and R, or R and r.
        """
        first_dim_nodes = self.mesh.nodes
        first_dim_edges = self.mesh.edges
        second_dim_nodes = self.base_variables[0].secondary_mesh.nodes
        second_dim_edges = self.base_variables[0].secondary_mesh.edges
        if self.base_eval_size // len(second_dim_nodes) == len(first_dim_nodes):
            first_dim_pts = first_dim_nodes
        elif self.base_eval_size // len(second_dim_nodes) == len(first_dim_edges):
            first_dim_pts = first_dim_edges

        second_dim_pts = second_dim_nodes
        first_dim_size = len(first_dim_pts)
        second_dim_size = len(second_dim_pts)

        entries = self.unroll_2D(
            realdata=None,
            n_dim1=second_dim_size,
            n_dim2=first_dim_size,
            axis_swaps=[(0, 2), (0, 1)],
        )

        # add points outside first dimension domain for extrapolation to
        # boundaries
        extrap_space_first_dim_left = np.array(
            [2 * first_dim_pts[0] - first_dim_pts[1]]
        )
        extrap_space_first_dim_right = np.array(
            [2 * first_dim_pts[-1] - first_dim_pts[-2]]
        )
        first_dim_pts = np.concatenate(
            [extrap_space_first_dim_left, first_dim_pts, extrap_space_first_dim_right]
        )
        extrap_entries_left = np.expand_dims(2 * entries[0] - entries[1], axis=0)
        extrap_entries_right = np.expand_dims(2 * entries[-1] - entries[-2], axis=0)
        entries_for_interp = np.concatenate(
            [extrap_entries_left, entries, extrap_entries_right], axis=0
        )

        # add points outside second dimension domain for extrapolation to
        # boundaries
        extrap_space_second_dim_left = np.array(
            [2 * second_dim_pts[0] - second_dim_pts[1]]
        )
        extrap_space_second_dim_right = np.array(
            [2 * second_dim_pts[-1] - second_dim_pts[-2]]
        )
        second_dim_pts = np.concatenate(
            [
                extrap_space_second_dim_left,
                second_dim_pts,
                extrap_space_second_dim_right,
            ]
        )
        extrap_entries_second_dim_left = np.expand_dims(
            2 * entries_for_interp[:, 0, :] - entries_for_interp[:, 1, :], axis=1
        )
        extrap_entries_second_dim_right = np.expand_dims(
            2 * entries_for_interp[:, -1, :] - entries_for_interp[:, -2, :], axis=1
        )
        entries_for_interp = np.concatenate(
            [
                extrap_entries_second_dim_left,
                entries_for_interp,
                extrap_entries_second_dim_right,
            ],
            axis=1,
        )

        # Process r-x, x-z, r-R, R-x, or R-z
        if self.domain[0].endswith("particle") and self.domains["secondary"][
            0
        ].endswith("electrode"):
            self.first_dimension = "r"
            self.second_dimension = "x"
            self.r_sol = first_dim_pts
            self.x_sol = second_dim_pts
        elif self.domain[0] in [
            "negative electrode",
            "separator",
            "positive electrode",
        ] and self.domains["secondary"] == ["current collector"]:
            self.first_dimension = "x"
            self.second_dimension = "z"
            self.x_sol = first_dim_pts
            self.z_sol = second_dim_pts
        elif self.domain[0].endswith("particle") and self.domains["secondary"][
            0
        ].endswith("particle size"):
            self.first_dimension = "r"
            self.second_dimension = "R"
            self.r_sol = first_dim_pts
            self.R_sol = second_dim_pts
        elif self.domain[0].endswith("particle size") and self.domains["secondary"][
            0
        ].endswith("electrode"):
            self.first_dimension = "R"
            self.second_dimension = "x"
            self.R_sol = first_dim_pts
            self.x_sol = second_dim_pts
        elif self.domain[0].endswith("particle size") and self.domains["secondary"] == [
            "current collector"
        ]:
            self.first_dimension = "R"
            self.second_dimension = "z"
            self.R_sol = first_dim_pts
            self.z_sol = second_dim_pts
        else:  # pragma: no cover
            raise pybamm.DomainError(
                f"Cannot process 2D object with domains '{self.domains}'."
            )

        # assign attributes for reference
        self.entries = entries
        self.dimensions = 2
        first_dim_pts_for_interp = first_dim_pts
        second_dim_pts_for_interp = second_dim_pts

        # Set pts to edges for nicer plotting
        self.first_dim_pts = first_dim_edges
        self.second_dim_pts = second_dim_edges

        # set up interpolation
        self._xr_data_array = xr.DataArray(
            entries_for_interp,
            coords={
                self.first_dimension: first_dim_pts_for_interp,
                self.second_dimension: second_dim_pts_for_interp,
                "t": self.t_pts,
            },
        )

    def initialise_2D_scikit_fem(self):
        y_sol = self.mesh.edges["y"]
        len_y = len(y_sol)
        z_sol = self.mesh.edges["z"]
        len_z = len(z_sol)
        entries = self.unroll_2D(
            realdata=None,
            n_dim1=len_y,
            n_dim2=len_z,
            axis_swaps=[(0, 2)],
        )

        # assign attributes for reference
        self.entries = entries
        self.dimensions = 2
        self.y_sol = y_sol
        self.z_sol = z_sol
        self.first_dimension = "y"
        self.second_dimension = "z"
        self.first_dim_pts = y_sol
        self.second_dim_pts = z_sol

        # set up interpolation
        self._xr_data_array = xr.DataArray(
            entries,
            coords={"y": y_sol, "z": z_sol, "t": self.t_pts},
        )

    def initialise_3D(self):
        """
        Initialise a 3D object that depends on x, r, and R.
        """
        first_dim_nodes = self.mesh.nodes
        first_dim_edges = self.mesh.edges
        second_dim_nodes = self.base_variables[0].secondary_mesh.nodes
        second_dim_edges = self.base_variables[0].secondary_mesh.edges
        third_dim_nodes = self.base_variables[0].tertiary_mesh.nodes
        third_dim_edges = self.base_variables[0].tertiary_mesh.edges
        if self.base_eval_size // (len(second_dim_nodes) * len(third_dim_nodes)) == len(
            first_dim_nodes
        ):
            first_dim_pts = first_dim_nodes
        elif self.base_eval_size // (
            len(second_dim_nodes) * len(third_dim_nodes)
        ) == len(first_dim_edges):
            first_dim_pts = first_dim_edges

        second_dim_pts = second_dim_nodes
        third_dim_pts = third_dim_nodes

        first_dim_size = len(first_dim_pts)
        second_dim_size = len(second_dim_pts)
        third_dim_size = len(third_dim_pts)

        entries = self.unroll_3D(
            realdata=None,
            n_dim1=third_dim_size,
            n_dim2=second_dim_size,
            n_dim3=first_dim_size,
            axis_swaps=[(0, 3), (0, 2), (0, 1)],
        )

        # add points outside first dimension domain for extrapolation to
        # boundaries
        extrap_space_first_dim_left = np.array(
            [2 * first_dim_pts[0] - first_dim_pts[1]]
        )
        extrap_space_first_dim_right = np.array(
            [2 * first_dim_pts[-1] - first_dim_pts[-2]]
        )
        first_dim_pts = np.concatenate(
            [extrap_space_first_dim_left, first_dim_pts, extrap_space_first_dim_right]
        )
        extrap_entries_left = np.expand_dims(2 * entries[0] - entries[1], axis=0)
        extrap_entries_right = np.expand_dims(2 * entries[-1] - entries[-2], axis=0)
        entries_for_interp = np.concatenate(
            [extrap_entries_left, entries, extrap_entries_right], axis=0
        )

        # add points outside second dimension domain for extrapolation to
        # boundaries
        extrap_space_second_dim_left = np.array(
            [2 * second_dim_pts[0] - second_dim_pts[1]]
        )
        extrap_space_second_dim_right = np.array(
            [2 * second_dim_pts[-1] - second_dim_pts[-2]]
        )
        second_dim_pts = np.concatenate(
            [
                extrap_space_second_dim_left,
                second_dim_pts,
                extrap_space_second_dim_right,
            ]
        )
        extrap_entries_second_dim_left = np.expand_dims(
            2 * entries_for_interp[:, 0, :] - entries_for_interp[:, 1, :], axis=1
        )
        extrap_entries_second_dim_right = np.expand_dims(
            2 * entries_for_interp[:, -1, :] - entries_for_interp[:, -2, :], axis=1
        )
        entries_for_interp = np.concatenate(
            [
                extrap_entries_second_dim_left,
                entries_for_interp,
                extrap_entries_second_dim_right,
            ],
            axis=1,
        )

        # add points outside tertiary dimension domain for extrapolation to
        # boundaries
        extrap_space_third_dim_left = np.array(
            [2 * third_dim_pts[0] - third_dim_pts[1]]
        )
        extrap_space_third_dim_right = np.array(
            [2 * third_dim_pts[-1] - third_dim_pts[-2]]
        )
        third_dim_pts = np.concatenate(
            [
                extrap_space_third_dim_left,
                third_dim_pts,
                extrap_space_third_dim_right,
            ]
        )
        extrap_entries_third_dim_left = np.expand_dims(
            2 * entries_for_interp[:, :, 0] - entries_for_interp[:, :, 1], axis=2
        )
        extrap_entries_third_dim_right = np.expand_dims(
            2 * entries_for_interp[:, :, -1] - entries_for_interp[:, :, -2], axis=2
        )
        entries_for_interp = np.concatenate(
            [
                extrap_entries_third_dim_left,
                entries_for_interp,
                extrap_entries_third_dim_right,
            ],
            axis=2,
        )

        # Process r-R-x only
        if (
            self.domain[0].endswith("particle")
            and self.domains["secondary"][0].endswith("particle size")
            and self.domains["tertiary"][0].endswith("electrode")
        ):
            self.first_dimension = "r"
            self.second_dimension = "R"
            self.third_dimension = "x"
            self.r_sol = first_dim_pts
            self.R_sol = second_dim_pts
            self.x_sol = third_dim_pts
        else:  # pragma: no cover
            raise pybamm.DomainError(
                f"Cannot process 3D object with domains '{self.domains}'."
            )

        # assign attributes for reference
        self.entries = entries
        self.dimensions = 3
        first_dim_pts_for_interp = first_dim_pts
        second_dim_pts_for_interp = second_dim_pts
        third_dim_pts_for_interp = third_dim_pts

        # Set pts to edges for nicer plotting
        self.first_dim_pts = first_dim_edges
        self.second_dim_pts = second_dim_edges
        self.third_dim_pts = third_dim_edges

        # set up interpolation
        self._xr_data_array = xr.DataArray(
            entries_for_interp,
            coords={
                self.first_dimension: first_dim_pts_for_interp,
                self.second_dimension: second_dim_pts_for_interp,
                self.third_dimension: third_dim_pts_for_interp,
                "t": self.t_pts,
            },
        )

    def initialise_3D_scikit_fem(self):
        x_nodes = self.mesh.nodes
        x_edges = self.mesh.edges
        y_sol = self.base_variables[0].secondary_mesh.edges["y"]
        z_sol = self.base_variables[0].secondary_mesh.edges["z"]
        if self.base_eval_size // (len(y_sol) * len(z_sol)) == len(x_nodes):
            x_sol = x_nodes
        elif self.base_eval_size // (len(y_sol) * len(z_sol)) == len(x_edges):
            x_sol = x_edges

        len_x = len(x_sol)
        len_y = len(y_sol)
        len_z = len(z_sol)
        entries = self.unroll_3D(
            realdata=None,
            n_dim1=len_z,
            n_dim2=len_y,
            n_dim3=len_x,
            axis_swaps=[(0, 3), (0, 2), (0, 1)],
        )

        # assign attributes for reference
        self.entries = entries
        self.dimensions = 3
        self.x_sol = x_sol
        self.y_sol = y_sol
        self.z_sol = z_sol
        self.first_dimension = "x"
        self.second_dimension = "y"
        self.third_dimension = "z"
        self.first_dim_pts = x_sol
        self.second_dim_pts = y_sol
        self.third_dim_pts = z_sol

        # set up interpolation
        self._xr_data_array = xr.DataArray(
            entries,
            coords={"x": x_sol, "y": y_sol, "z": z_sol, "t": self.t_pts},
        )

    def __call__(self, t=None, x=None, r=None, y=None, z=None, R=None):
        """
        Evaluate the variable at arbitrary *dimensional* t (and x, r, y, z and/or R),
        using interpolation
        """
        if self.time_indep:
            return self.entries
        kwargs = {"t": t, "x": x, "r": r, "y": y, "z": z, "R": R}
        # Remove any None arguments
        kwargs = {key: value for key, value in kwargs.items() if value is not None}
        # Use xarray interpolation, return numpy array
        return self._xr_data_array.interp(**kwargs).values

    @property
    def data(self):
        """Same as entries, but different name"""
        return self.entries

    @property
    def sensitivities(self):
        """
        Returns a dictionary of sensitivities for each input parameter.
        The keys are the input parameters, and the value is a matrix of size
        (n_x * n_t, n_p), where n_x is the number of states, n_t is the number of time
        points, and n_p is the size of the input parameter
        """
        # No sensitivities if there are no inputs
        if len(self.all_inputs[0]) == 0:
            return {}
        return self._sensitivities

    def _concat(
        self, other: pybamm.ProcessedVariableComputed, new_sol: pybamm.Solution
    ) -> pybamm.ProcessedVariableComputed:
        """
        Returns a new ProcessedVariableComputed object that is the result of appending
        the data from other to this object. Used exclusively in running experiments, to
        append the data from one cycle to the next.

        Parameters
        ----------
        other : :class:`pybamm.ProcessedVariableComputed`
            The other ProcessedVariableComputed object to append to this one
        new_sol : :class:`pybamm.Solution`
            The new solution object to be used to create the processed variables
        """

        bv = self.base_variables + other.base_variables
        bvc = self.base_variables_casadi + other.base_variables_casadi
        bvd = self.base_variables_data + other.base_variables_data

        new_var = self.__class__(bv, bvc, bvd, new_sol)

        new_var._sensitivities = {
            k: np.concatenate((self.sensitivities[k], other.sensitivities[k]))
            for k in self.sensitivities.keys()
        }

        return new_var



#
# Base class for SEI models.
#
import pybamm
from pybamm.models.submodels.interface.base_interface import BaseInterface


class BaseModel(BaseInterface):
    """Base class for SEI models.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict
        A dictionary of options to be passed to the model.
    phase : str, optional
        Phase of the particle (default is "primary")
    cracks : bool, optional
        Whether this is a submodel for standard SEI or SEI on cracks
    """

    def __init__(self, param, domain, options, phase="primary", cracks=False):
        if cracks is True:
            reaction = "SEI on cracks"
        else:
            reaction = "SEI"
        super().__init__(param, domain, reaction, options=options, phase=phase)

    def get_coupled_variables(self, variables):
        # Update some common variables
        domain, Domain = self.domain_Domain

        if self.reaction_loc != "interface":
            j_sei_av = variables[
                f"X-averaged {domain} electrode {self.reaction_name}interfacial"
                " current density [A.m-2]"
            ]
            j_sei = variables[
                f"{Domain} electrode {self.reaction_name}interfacial current"
                " density [A.m-2]"
            ]
            variables.update(
                {
                    f"X-averaged {domain} electrode {self.reaction_name}interfacial "
                    "current density [A.m-2]": j_sei_av,
                    f"{Domain} electrode {self.reaction_name}interfacial current "
                    "density [A.m-2]": j_sei,
                }
            )

        variables.update(
            self._get_standard_volumetric_current_density_variables(variables)
        )

        return variables

    def _get_standard_concentration_variables(self, c_sei):
        """
        A private function to obtain the standard variables which
        can be derived from the local SEI concentration.

        Parameters
        ----------
        c : :class:`pybamm.Symbol`
            The SEI concentration.

        Returns
        -------
        variables : dict
            The variables which can be derived from the SEI concentration.
        """
        domain, Domain = self.domain_Domain
        phase_param = self.phase_param
        reaction_name = self.reaction_name

        if self.reaction_loc == "interface":
            # c_sei is an interfacial quantity [mol.m-2]
            c_sei_av = pybamm.yz_average(c_sei)
            variables = {
                f"{Domain} {reaction_name}concentration [mol.m-2]": c_sei,
                f"Volume-averaged {domain} {reaction_name}concentration "
                "[mol.m-2]": c_sei_av,
            }
        else:
            # c_sei is a bulk quantity [mol.m-3]
            if self.size_distribution:
                c_sei_sav = pybamm.size_average(c_sei)
                c_sei_xav = pybamm.x_average(c_sei_sav)
            else:
                c_sei_xav = pybamm.x_average(c_sei)
            c_sei_av = pybamm.yz_average(c_sei_xav)
            variables = {
                f"{Domain} {self.reaction_name}concentration [mol.m-3]": c_sei,
                f"X-averaged {domain} {reaction_name}concentration "
                "[mol.m-3]": c_sei_xav,
                f"Volume-averaged {domain} {reaction_name}concentration "
                "[mol.m-3]": c_sei_av,
            }

        # Set scales to one for the "no SEI" model so that they are not required
        # by parameter values in general
        if isinstance(self, pybamm.sei.NoSEI):
            c_sei_0 = 0
            c_sei_cr0 = 0
            z_sei = 1
        else:
            L_sei_0 = phase_param.L_sei_0
            V_bar_sei = phase_param.V_bar_sei
            z_sei = phase_param.z_sei
            if self.reaction_loc == "interface":
                c_sei_0 = L_sei_0 / V_bar_sei  # mol.m-2
            else:
                c_sei_0 = L_sei_0 * phase_param.a_typ / V_bar_sei  # mol.m-3
                c_sei_cr0 = phase_param.L_sei_cr0 * phase_param.a_typ / V_bar_sei

        if self.reaction == "SEI":
            delta_c_sei = c_sei_av - c_sei_0
        elif self.reaction == "SEI on cracks":
            delta_c_sei = c_sei_av - c_sei_cr0

        if self.reaction_loc == "interface":
            L_k = 1
        elif domain == "negative":
            L_k = self.param.n.L
        elif domain == "positive":
            L_k = self.param.p.L

        # Multiply delta_n_SEI by V_k to get total moles of SEI formed
        # Multiply by z_sei to get total lithium moles consumed by SEI
        V_k = L_k * self.param.L_y * self.param.L_z
        Q_sei = delta_c_sei * V_k * z_sei

        variables.update(
            {
                f"Loss of lithium to {domain} {self.reaction_name}[mol]": Q_sei,
                f"Loss of capacity to {domain} {self.reaction_name}[A.h]": Q_sei
                * self.param.F
                / 3600,
            }
        )

        return variables

    def _get_standard_reaction_variables(self, j_sei):
        """
        A private function to obtain the standard variables which
        can be derived from the SEI interfacial reaction current

        Parameters
        ----------
        j_sei : :class:`pybamm.Symbol`
            The SEI interfacial reaction current.

        Returns
        -------
        variables : dict
            The variables which can be derived from the SEI currents.
        """
        domain, Domain = self.domain_Domain

        variables = {
            f"{Domain} electrode {self.reaction_name}"
            "interfacial current density [A.m-2]": j_sei,
        }

        if self.reaction_loc != "interface":
            j_sei_av = pybamm.x_average(j_sei)
            variables.update(
                {
                    f"X-averaged {domain} electrode {self.reaction_name}"
                    "interfacial current density [A.m-2]": j_sei_av,
                }
            )

        return variables

    def _get_standard_reaction_distribution_variables(self, j_sei):
        """
        A private function to obtain the standard variables which
        can be derived from the SEI interfacial reaction current

        Parameters
        ----------
        j_sei : :class:`pybamm.Symbol`
            The SEI interfacial reaction current.

        Returns
        -------
        variables : dict
            The variables which can be derived from the SEI currents.
        """
        domain, Domain = self.domain_Domain
        j_sei = pybamm.size_average(j_sei)
        variables = {
            f"{Domain} electrode {self.reaction_name}"
            "interfacial current density [A.m-2]": j_sei,
        }

        if self.reaction_loc != "interface":
            j_sei_av = pybamm.x_average(j_sei)
            variables.update(
                {
                    f"X-averaged {domain} electrode {self.reaction_name}"
                    "interfacial current density [A.m-2]": j_sei_av,
                }
            )

        return variables



#
# Solution class
#
import json
import numbers
import pickle
from functools import cached_property

import casadi
import numpy as np
import pandas as pd
from scipy.io import savemat

import pybamm


class NumpyEncoder(json.JSONEncoder):
    """
    Numpy serialiser helper class that converts numpy arrays to a list.
    Numpy arrays cannot be directly converted to JSON, so the arrays are
    converted to python list objects before encoding.
    """

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        # won't be called since we only need to convert numpy arrays
        return json.JSONEncoder.default(self, obj)  # pragma: no cover


class Solution:
    """
    Class containing the solution of, and various attributes associated with, a PyBaMM
    model.

    Parameters
    ----------
    all_ts : :class:`numpy.array`, size (n,) (or list of these)
        A one-dimensional array containing the times at which the solution is evaluated.
        A list of times can be provided instead to initialize a solution with
        sub-solutions.
    all_ys : :class:`numpy.array`, size (m, n) (or list of these)
        A two-dimensional array containing the values of the solution. y[i, :] is the
        vector of solutions at time t[i].
        A list of ys can be provided instead to initialize a solution with
        sub-solutions.
    all_models : :class:`pybamm.BaseModel`
        The model that was used to calculate the solution.
        A list of models can be provided instead to initialize a solution with
        sub-solutions that have been calculated using those models.
    all_inputs : dict (or list of these)
        The inputs that were used to calculate the solution
        A list of inputs can be provided instead to initialize a solution with
        sub-solutions.
    t_event : :class:`numpy.array`, size (1,)
        A zero-dimensional array containing the time at which the event happens.
    y_event : :class:`numpy.array`, size (m,)
        A one-dimensional array containing the value of the solution at the time when
        the event happens.
    termination : str
        String to indicate why the solution terminated
    all_sensitivities: dict of lists
        sensitivities are provided as a dict of {parameter: [sensitivities]} pairs.
    variables_returned: bool
        Bool to indicate if `all_ys` contains the full state vector, or is empty because
        only requested variables have been returned. True if `output_variables` is used
        with a solver, otherwise False.

    """

    def __init__(
        self,
        all_ts,
        all_ys,
        all_models,
        all_inputs,
        t_event=None,
        y_event=None,
        termination="final time",
        all_sensitivities=None,
        all_yps=None,
        variables_returned=False,
        check_solution=True,
    ):
        if not isinstance(all_ts, list):
            all_ts = [all_ts]
        if not isinstance(all_ys, list):
            all_ys = [all_ys]
        if not isinstance(all_models, list):
            all_models = [all_models]
        self._all_ts = all_ts
        self._all_ys = all_ys
        self._all_ys_and_sens = all_ys
        self._all_models = all_models

        if (all_yps is not None) and not isinstance(all_yps, list):
            all_yps = [all_yps]
        self._all_yps = all_yps

        self.variables_returned = variables_returned

        # Set up inputs
        if not isinstance(all_inputs, list):
            all_inputs_copy = dict(all_inputs)
            for key, value in all_inputs_copy.items():
                if isinstance(value, numbers.Number):
                    all_inputs_copy[key] = np.array([value])
            self.all_inputs = [all_inputs_copy]
        else:
            self.all_inputs = all_inputs

        if all_sensitivities is None:
            all_sensitivities = {}
        if not isinstance(all_sensitivities, dict):
            raise TypeError(
                "sensitivities arg needs to be a dict of {parameter: [sensitivities]} "
            )
        self._all_sensitivities = {}
        for key, value in all_sensitivities.items():
            if isinstance(value, list):
                self._all_sensitivities[key] = value
            else:
                self._all_sensitivities[key] = [value]

        # Check no ys are too large
        if check_solution:
            self.check_ys_are_not_too_large()

        # Events
        self._t_event = t_event
        self._y_event = y_event
        self._termination = termination
        self.closest_event_idx = None

        # Initialize times
        self.set_up_time = None
        self.solve_time = None
        self.integration_time = None

        # initialize empty variables and data
        self._variables = pybamm.FuzzyDict()
        self._data = pybamm.FuzzyDict()

        # Add self as sub-solution for compatibility with ProcessedVariable
        self._sub_solutions = [self]

        # initialize empty cycles
        self._cycles = []

        # Initialize empty summary variables
        self._summary_variables = None

        # Initialise initial start time
        self.initial_start_time = None

        # Solution now uses CasADi
        pybamm.citations.register("Andersson2019")

    def has_sensitivities(self) -> bool:
        return len(self._all_sensitivities) > 0

    @property
    def t(self):
        """Times at which the solution is evaluated"""
        try:
            return self._t
        except AttributeError:
            self.set_t()
            return self._t

    def set_t(self):
        self._t = np.concatenate(self.all_ts)
        if any(np.diff(self._t) <= 0):
            raise ValueError("Solution time vector must be strictly increasing")

    @property
    def y(self):
        """Values of the solution"""
        try:
            return self._y
        except AttributeError:
            self.set_y()

            return self._y

    @property
    def data(self):
        for k, v in self._variables.items():
            if k not in self._data:
                self._data[k] = v.data
        return self._data

    @property
    def sensitivities(self):
        """Values of the sensitivities. Returns a dict of param_name: np_array"""
        try:
            return self._sensitivities
        except AttributeError:
            self.set_sensitivities()
        return self._sensitivities

    def set_sensitivities(self):
        if not self.has_sensitivities():
            self._sensitivities = {}
            return

        self._sensitivities = {}
        for key, sens in self._all_sensitivities.items():
            self._sensitivities[key] = np.vstack(sens)

    def set_y(self):
        try:
            if isinstance(self.all_ys[0], casadi.DM | casadi.MX):
                self._y = casadi.horzcat(*self.all_ys)
            else:
                self._y = np.hstack(self.all_ys)
        except ValueError as error:
            raise pybamm.SolverError(
                "The solution is made up from different models, so `y` cannot be "
                "computed explicitly."
            ) from error

    def check_ys_are_not_too_large(self):
        # Only check last one so that it doesn't take too long
        # We only care about the cases where y is growing too large without any
        # restraint, so if y gets large in the middle then comes back down that is ok
        y, model = self.all_ys[-1], self.all_models[-1]
        y = y[:, -1]
        if np.any(y > pybamm.settings.max_y_value):
            for var in [*model.rhs.keys(), *model.algebraic.keys()]:
                var = model.variables[var.name]
                # find the statevector corresponding to this variable
                statevector = None
                for node in var.pre_order():
                    if isinstance(node, pybamm.StateVector):
                        statevector = node

                # there will always be a statevector, but just in case
                if statevector is None:  # pragma: no cover
                    raise RuntimeError(
                        f"Cannot find statevector corresponding to variable {var.name}"
                    )
                y_var = y[statevector.y_slices[0]]
                if np.any(y_var > pybamm.settings.max_y_value):
                    pybamm.logger.error(
                        f"Solution for '{var}' exceeds the maximum allowed value "
                        f"of `{pybamm.settings.max_y_value}. This could be due to "
                        "incorrect scaling, model formulation, or "
                        "parameter values. The maximum allowed value is set by "
                        "'pybammm.settings.max_y_value'."
                    )

    @property
    def all_ts(self):
        return self._all_ts

    @property
    def all_ys(self):
        return self._all_ys

    @property
    def all_models(self):
        """Model(s) used for solution"""
        return self._all_models

    @cached_property
    def all_inputs_casadi(self):
        return [casadi.vertcat(*inp.values()) for inp in self.all_inputs]

    @property
    def all_yps(self):
        return self._all_yps

    @property
    def hermite_interpolation(self):
        return self.all_yps is not None

    @property
    def t_event(self):
        """Time at which the event happens"""
        return self._t_event

    @property
    def y_event(self):
        """Value of the solution at the time of the event"""
        return self._y_event

    @property
    def termination(self):
        """Reason for termination"""
        return self._termination

    @termination.setter
    def termination(self, value):
        """Updates the reason for termination"""
        self._termination = value

    @cached_property
    def first_state(self):
        """
        A Solution object that only contains the first state. This is faster to evaluate
        than the full solution when only the first state is needed (e.g. to initialize
        a model with the solution)
        """
        sensitivities = {}
        n_states = self.all_models[0].len_rhs_and_alg
        for key in self._all_sensitivities:
            sensitivities[key] = self._all_sensitivities[key][0][-n_states:, :]

        if self.all_yps is None:
            all_yps = None
        else:
            all_yps = self.all_yps[0][:, :1]

        if not self.variables_returned:
            all_ys = self.all_ys[0][:, :1]
        else:
            # Get first state from initial conditions as all_ys is empty
            all_ys = self.all_models[0].y0full.reshape(-1, 1)

        new_sol = Solution(
            self.all_ts[0][:1],
            all_ys,
            self.all_models[:1],
            self.all_inputs[:1],
            None,
            None,
            "final time",
            all_sensitivities=sensitivities,
            all_yps=all_yps,
        )
        new_sol._all_inputs_casadi = self.all_inputs_casadi[:1]
        new_sol._sub_solutions = self.sub_solutions[:1]

        new_sol.solve_time = 0
        new_sol.integration_time = 0
        new_sol.set_up_time = 0

        return new_sol

    @cached_property
    def last_state(self):
        """
        A Solution object that only contains the final state. This is faster to evaluate
        than the full solution when only the final state is needed (e.g. to initialize
        a model with the solution)
        """
        sensitivities = {}
        n_states = self.all_models[-1].len_rhs_and_alg
        for key in self._all_sensitivities:
            sensitivities[key] = self._all_sensitivities[key][-1][-n_states:, :]

        if self.all_yps is None:
            all_yps = None
        else:
            all_yps = self.all_yps[-1][:, -1:]

        if not self.variables_returned:
            all_ys = self.all_ys[-1][:, -1:]
        else:
            # Get last state from y_event as all_ys is empty
            all_ys = self.y_event.reshape(len(self.y_event), 1)

        new_sol = Solution(
            self.all_ts[-1][-1:],
            all_ys,
            self.all_models[-1:],
            self.all_inputs[-1:],
            self.t_event,
            self.y_event,
            self.termination,
            all_sensitivities=sensitivities,
            all_yps=all_yps,
        )
        new_sol._all_inputs_casadi = self.all_inputs_casadi[-1:]
        new_sol._sub_solutions = self.sub_solutions[-1:]
        new_sol.solve_time = 0
        new_sol.integration_time = 0
        new_sol.set_up_time = 0

        return new_sol

    @property
    def total_time(self):
        return self.set_up_time + self.solve_time

    @property
    def cycles(self):
        return self._cycles

    @cycles.setter
    def cycles(self, cycles):
        self._cycles = cycles

    @property
    def summary_variables(self):
        return self._summary_variables

    @property
    def initial_start_time(self):
        return self._initial_start_time

    @initial_start_time.setter
    def initial_start_time(self, value):
        """Updates the initial start time of the experiment"""
        self._initial_start_time = value

    def update_summary_variables(self, all_summary_variables):
        self.all_summary_variables = all_summary_variables
        self._summary_variables = pybamm.SummaryVariables(
            self, cycle_summary_variables=all_summary_variables
        )

    def update(self, variables):
        """Add ProcessedVariables to the dictionary of variables in the solution"""
        # Single variable
        if isinstance(variables, str):
            variables = [variables]

        # Process
        for variable in variables:
            self._update_variable(variable)

    def _update_variable(self, variable):
        time_integral = None
        pybamm.logger.debug(f"Post-processing {variable}")
        vars_pybamm = [
            model.variables_and_events[variable] for model in self.all_models
        ]

        # Iterate through all models, some may be in the list several times and
        # therefore only get set up once
        vars_casadi = []
        for i, (model, ys, inputs, var_pybamm) in enumerate(
            zip(self.all_models, self.all_ys, self.all_inputs, vars_pybamm, strict=True)
        ):
            if self.variables_returned and var_pybamm.has_symbol_of_classes(
                pybamm.expression_tree.state_vector.StateVector
            ):
                raise KeyError(
                    f"Cannot process variable '{variable}' as it was not part of the "
                    "solve. Please re-run the solve with `output_variables` set to "
                    "include this variable."
                )
            elif variable in model._variables_casadi:
                var_casadi = model._variables_casadi[variable]
            else:
                time_integral = pybamm.ProcessedVariableTimeIntegral.from_pybamm_var(
                    var_pybamm, self.all_ys[i].shape[0]
                )
                if time_integral is not None:
                    vars_pybamm[i] = time_integral.sum_node.child
                    var_casadi = self.process_casadi_var(
                        time_integral.sum_node.child,
                        inputs,
                        ys.shape,
                    )
                    if time_integral.post_sum_node is not None:
                        time_integral.post_sum = self.process_casadi_var(
                            time_integral.post_sum_node,
                            inputs,
                            ys.shape,
                        )
                else:
                    var_casadi = self.process_casadi_var(
                        var_pybamm,
                        inputs,
                        ys.shape,
                    )
                    model._variables_casadi[variable] = var_casadi
            vars_casadi.append(var_casadi)
        var = pybamm.process_variable(
            variable, vars_pybamm, vars_casadi, self, time_integral=time_integral
        )

        self._variables[variable] = var

    def process_casadi_var(self, var_pybamm, inputs, ys_shape):
        t_MX = casadi.MX.sym("t")
        y_MX = casadi.MX.sym("y", ys_shape[0])
        inputs_MX_dict = {
            key: casadi.MX.sym("input", value.shape[0]) for key, value in inputs.items()
        }
        inputs_MX = casadi.vertcat(*[p for p in inputs_MX_dict.values()])
        var_sym = var_pybamm.to_casadi(t_MX, y_MX, inputs=inputs_MX_dict)

        opts = {
            "cse": True,
            "inputs_check": False,
            "is_diff_in": [False, False, False],
            "is_diff_out": [False],
            "regularity_check": False,
            "error_on_fail": False,
            "enable_jacobian": False,
        }

        # Casadi has a bug where it does not correctly handle arrays with
        # zeros padded at the beginning or end. To avoid this, we add and
        # subtract the same number to the variable to reinforce the
        # variable bounds. This does not affect the answer
        epsilon = 1.0
        var_sym = (var_sym - epsilon) + epsilon

        var_casadi = casadi.Function(
            "variable",
            [t_MX, y_MX, inputs_MX],
            [var_sym],
            opts,
        )

        # Some variables, like interpolants, cannot be expanded
        try:
            var_casadi_out = var_casadi.expand()
        except RuntimeError as error:
            if "'eval_sx' not defined for" not in str(error):
                raise error  # pragma: no cover
            var_casadi_out = var_casadi

        return var_casadi_out

    def __getitem__(self, key):
        """Read a variable from the solution. Variables are created 'just in time', i.e.
        only when they are called.

        Parameters
        ----------
        key : str
            The name of the variable

        Returns
        -------
        :class:`pybamm.ProcessedVariable` or :class:`pybamm.ProcessedVariableComputed`
            A variable that can be evaluated at any time or spatial point. The
            underlying data for this variable is available in its attribute ".data"
        """

        # return it if it exists
        if key in self._variables:
            return self._variables[key]
        else:
            # otherwise create it, save it and then return it
            self.update(key)
            return self._variables[key]

    def plot(self, output_variables=None, **kwargs):
        """
        A method to quickly plot the outputs of the solution. Creates a
        :class:`pybamm.QuickPlot` object (with keyword arguments 'kwargs') and
        then calls :meth:`pybamm.QuickPlot.dynamic_plot`.

        Parameters
        ----------
        output_variables: list, optional
            A list of the variables to plot.
        **kwargs
            Additional keyword arguments passed to
            :meth:`pybamm.QuickPlot.dynamic_plot`.
            For a list of all possible keyword arguments see :class:`pybamm.QuickPlot`.
        """
        return pybamm.dynamic_plot(self, output_variables=output_variables, **kwargs)

    def save(self, filename):
        """Save the whole solution using pickle"""
        # No warning here if len(self.data)==0 as solution can be loaded
        # and used to process new variables

        with open(filename, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    def get_data_dict(self, variables=None, short_names=None, cycles_and_steps=True):
        """
        Construct a (standard python) dictionary of the solution data containing the
        variables in `variables`. If `variables` is None then all variables are
        returned. Any variable names in short_names are replaced with the corresponding
        short name.

        If the solution has cycles, then the cycle numbers and step numbers are also
        returned in the dictionary.

        Parameters
        ----------
        variables : list, optional
            List of variables to return. If None, returns all variables in solution.data
        short_names : dict, optional
            Dictionary of shortened names to use when saving.
        cycles_and_steps : bool, optional
            Whether to include the cycle numbers and step numbers in the dictionary

        Returns
        -------
        dict
            A dictionary of the solution data
        """
        if variables is None:
            # variables not explicitly provided -> save all variables that have been
            # computed
            data_long_names = self.data
        else:
            if isinstance(variables, str):
                variables = [variables]
            # otherwise, save only the variables specified
            data_long_names = {}
            for name in variables:
                data_long_names[name] = self[name].data
        if len(data_long_names) == 0:
            raise ValueError(
                """
                Solution does not have any data. Please provide a list of variables
                to save.
                """
            )

        # Use any short names if provided
        data_short_names = {}
        short_names = short_names or {}
        for name, var in data_long_names.items():
            name = short_names.get(name, name)  # return name if no short name
            data_short_names[name] = var

        # Save cycle number and step number if the solution has them
        if cycles_and_steps and len(self.cycles) > 0:
            data_short_names["Cycle"] = np.array([])
            data_short_names["Step"] = np.array([])
            for i, cycle in enumerate(self.cycles):
                data_short_names["Cycle"] = np.concatenate(
                    [data_short_names["Cycle"], i * np.ones_like(cycle.t)]
                )
                for j, step in enumerate(cycle.steps):
                    data_short_names["Step"] = np.concatenate(
                        [data_short_names["Step"], j * np.ones_like(step.t)]
                    )

        return data_short_names

    def save_data(
        self, filename=None, variables=None, to_format="pickle", short_names=None
    ):
        """
        Save solution data only (raw arrays)

        Parameters
        ----------
        filename : str, optional
            The name of the file to save data to. If None, then a str is returned
        variables : list, optional
            List of variables to save. If None, saves all of the variables that have
            been created so far
        to_format : str, optional
            The format to save to. Options are:

            - 'pickle' (default): creates a pickle file with the data dictionary
            - 'matlab': creates a .mat file, for loading in matlab
            - 'csv': creates a csv file (0D variables only)
            - 'json': creates a json file
        short_names : dict, optional
            Dictionary of shortened names to use when saving. This may be necessary when
            saving to MATLAB, since no spaces or special characters are allowed in
            MATLAB variable names. Note that not all the variables need to be given
            a short name.

        Returns
        -------
        data : str, optional
            str if 'csv' or 'json' is chosen and filename is None, otherwise None
        """
        data = self.get_data_dict(variables=variables, short_names=short_names)

        if to_format == "pickle":
            if filename is None:
                raise ValueError("pickle format must be written to a file")
            with open(filename, "wb") as f:
                pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)
        elif to_format == "matlab":
            if filename is None:
                raise ValueError("matlab format must be written to a file")
            # Check all the variable names only contain a-z, A-Z or _ or numbers
            for name in data.keys():
                # Check the string only contains the following ASCII:
                # a-z (97-122)
                # A-Z (65-90)
                # _ (95)
                # 0-9 (48-57) but not in the first position
                for i, s in enumerate(name):
                    if not (
                        97 <= ord(s) <= 122
                        or 65 <= ord(s) <= 90
                        or ord(s) == 95
                        or (i > 0 and 48 <= ord(s) <= 57)
                    ):
                        raise ValueError(
                            f"Invalid character '{s}' found in '{name}'. "
                            + "MATLAB variable names must only contain a-z, A-Z, _, "
                            "or 0-9 (except the first position). "
                            "Use the 'short_names' argument to pass an alternative "
                            "variable name, e.g. \n\n"
                            "\tsolution.save_data(filename, "
                            "['Electrolyte concentration'], to_format='matlab, "
                            "short_names={'Electrolyte concentration': 'c_e'})"
                        )
            savemat(filename, data)
        elif to_format == "csv":
            for name, var in data.items():
                if var.ndim >= 2:
                    raise ValueError(
                        f"only 0D variables can be saved to csv, but '{name}' is {var.ndim - 1}D"
                    )
            df = pd.DataFrame(data)
            return df.to_csv(filename, index=False)
        elif to_format == "json":
            if filename is None:
                return json.dumps(data, cls=NumpyEncoder)
            else:
                with open(filename, "w") as outfile:
                    json.dump(data, outfile, cls=NumpyEncoder)
        else:
            raise ValueError(f"format '{to_format}' not recognised")

    @property
    def sub_solutions(self):
        """List of sub solutions that have been
        concatenated to form the full solution"""

        return self._sub_solutions

    def __add__(self, other):
        """Adds two solutions together, e.g. when stepping"""
        if other is None or isinstance(other, EmptySolution):
            return self.copy()
        if not isinstance(other, Solution):
            raise pybamm.SolverError(
                "Only a Solution or None can be added to a Solution"
            )
        # Special case: new solution only has one timestep and it is already in the
        # existing solution. In this case, return a copy of the existing solution
        if (
            len(other.all_ts) == 1
            and len(other.all_ts[0]) == 1
            and other.all_ts[0][0] == self.all_ts[-1][-1]
        ):
            new_sol = self.copy()
            # Update termination using the latter solution
            new_sol._termination = other.termination
            new_sol._t_event = other._t_event
            new_sol._y_event = other._y_event
            return new_sol

        # Update list of sub-solutions
        hermite_interpolation = (
            other.hermite_interpolation and self.hermite_interpolation
        )
        if other.all_ts[0][0] == self.all_ts[-1][-1]:
            # Skip first time step if it is repeated
            all_ts = [*self.all_ts, other.all_ts[0][1:], *other.all_ts[1:]]
            all_ys = [*self.all_ys, other.all_ys[0][:, 1:], *other.all_ys[1:]]
            if hermite_interpolation:
                all_yps = [*self.all_yps, other.all_yps[0][:, 1:], *other.all_yps[1:]]
        else:
            all_ts = self.all_ts + other.all_ts
            all_ys = self.all_ys + other.all_ys
            if hermite_interpolation:
                all_yps = self.all_yps + other.all_yps

        if not hermite_interpolation:
            all_yps = None

        # sensitivities are a dict of {parameter: [sensitivities]}
        # we can assume that the keys are the same for both solutions
        all_sensitivities = self._all_sensitivities
        for key in other._all_sensitivities:
            all_sensitivities[key] = (
                all_sensitivities[key] + other._all_sensitivities[key]
            )

        new_sol = Solution(
            all_ts,
            all_ys,
            self.all_models + other.all_models,
            self.all_inputs + other.all_inputs,
            other.t_event,
            other.y_event,
            other.termination,
            all_sensitivities=all_sensitivities,
            all_yps=all_yps,
            variables_returned=other.variables_returned,
        )

        new_sol.closest_event_idx = other.closest_event_idx
        new_sol._all_inputs_casadi = self.all_inputs_casadi + other.all_inputs_casadi

        # Add timers (if available)
        for attr in ["solve_time", "integration_time", "set_up_time"]:
            if (
                getattr(self, attr, None) is not None
                and getattr(other, attr, None) is not None
            ):
                setattr(new_sol, attr, getattr(self, attr) + getattr(other, attr))

        # Set sub_solutions
        new_sol._sub_solutions = self.sub_solutions + other.sub_solutions

        # update variables which were derived at the solver stage
        if any([self.variables_returned, other.variables_returned]):
            vars = {*self._variables.keys(), *other._variables.keys()}
            new_sol._variables = {v: self[v].update(other[v], new_sol) for v in vars}
            new_sol.variables_returned = True

        return new_sol

    def __radd__(self, other):
        return self.__add__(other)

    def copy(self):
        new_sol = self.__class__(
            self.all_ts,
            self.all_ys,
            self.all_models,
            self.all_inputs,
            self.t_event,
            self.y_event,
            self.termination,
            self._all_sensitivities,
            self.all_yps,
            self.variables_returned,
        )
        new_sol._all_inputs_casadi = self.all_inputs_casadi
        new_sol._sub_solutions = self.sub_solutions
        new_sol.closest_event_idx = self.closest_event_idx

        new_sol.solve_time = self.solve_time
        new_sol.integration_time = self.integration_time
        new_sol.set_up_time = self.set_up_time

        # copy over variables which were derived at the solver stage
        if self._variables and all(
            isinstance(v, pybamm.ProcessedVariableComputed)
            for v in self._variables.values()
        ):
            new_sol._variables = self._variables.copy()

        return new_sol

    def plot_voltage_components(
        self,
        ax=None,
        show_legend=True,
        split_by_electrode=False,
        show_plot=True,
        **kwargs_fill,
    ):
        """
        Generate a plot showing the component overpotentials that make up the voltage

        Parameters
        ----------
        ax : matplotlib Axis, optional
            The axis on which to put the plot. If None, a new figure and axis is created.
        show_legend : bool, optional
            Whether to display the legend. Default is True.
        split_by_electrode : bool, optional
            Whether to show the overpotentials for the negative and positive electrodes
            separately. Default is False.
        show_plot : bool, optional
            Whether to show the plots. Default is True. Set to False if you want to
            only display the plot after plt.show() has been called.
        kwargs_fill
            Keyword arguments, passed to ax.fill_between.

        """
        return pybamm.plot_voltage_components(
            self,
            ax=ax,
            show_legend=show_legend,
            split_by_electrode=split_by_electrode,
            show_plot=show_plot,
            **kwargs_fill,
        )


class EmptySolution:
    def __init__(self, termination=None, t=None):
        self.termination = termination
        if t is None:
            t = np.array([0])
        elif isinstance(t, numbers.Number):
            t = np.array([t])

        self.t = t

    def __add__(self, other):
        if isinstance(other, EmptySolution | Solution):
            return other.copy()

    def __radd__(self, other):
        if other is None:
            return self.copy()

    def copy(self):
        return EmptySolution(termination=self.termination, t=self.t)


def make_cycle_solution(
    step_solutions, esoh_solver=None, save_this_cycle=True, inputs=None
):
    """
    Function to create a Solution for an entire cycle, and associated summary variables

    Parameters
    ----------
    step_solutions : list of :class:`Solution`
        Step solutions that form the entire cycle
    esoh_solver : :class:`pybamm.lithium_ion.ElectrodeSOHSolver`
        Solver to calculate electrode SOH (eSOH) variables. If `None` (default)
        then only summary variables that do not require the eSOH calculation
        are calculated. See :footcite:t:`Mohtat2019` for more details on eSOH variables.
    save_this_cycle : bool, optional
        Whether to save the entire cycle variables or just the summary variables.
        Default True
    inputs : dict
        User inputs for the summary variables.

    Returns
    -------
    cycle_solution : :class:`pybamm.Solution` or None
        The Solution object for this cycle, or None (if save_this_cycle is False)
    cycle_summary_variables : dict
        Dictionary of summary variables for this cycle
    cycle_first_state : Solution
        First state of the cycle.

    """
    sum_sols = step_solutions[0].copy()
    for step_solution in step_solutions[1:]:
        sum_sols = sum_sols + step_solution

    cycle_solution = Solution(
        sum_sols.all_ts,
        sum_sols.all_ys,
        sum_sols.all_models,
        sum_sols.all_inputs,
        sum_sols.t_event,
        sum_sols.y_event,
        sum_sols.termination,
        sum_sols._all_sensitivities,
        sum_sols.all_yps,
        sum_sols.variables_returned,
    )

    if sum_sols.variables_returned:
        cycle_solution._variables = sum_sols._variables

    cycle_solution._all_inputs_casadi = sum_sols.all_inputs_casadi
    cycle_solution._sub_solutions = sum_sols.sub_solutions

    cycle_solution.solve_time = sum_sols.solve_time
    cycle_solution.integration_time = sum_sols.integration_time
    cycle_solution.set_up_time = sum_sols.set_up_time

    cycle_solution.steps = step_solutions

    cycle_summary_variables = pybamm.SummaryVariables(
        cycle_solution, esoh_solver=esoh_solver, user_inputs=inputs
    )

    cycle_first_state = cycle_solution.first_state

    if save_this_cycle:
        cycle_solution.cycle_summary_variables = cycle_summary_variables
    else:
        cycle_solution = None

    return cycle_solution, cycle_summary_variables, cycle_first_state
